@@ -3,6 +3,7 @@
 
 #include "Luau/Ast.h"
 #include "Luau/StringUtils.h"
+#include "Luau/Common.h"
 
 namespace Luau
 {
@@ -261,7 +262,7 @@ struct AstJsonEncoder : public AstVisitor
             if (comma)
                 writeRaw(",");
             else
-                comma = false;
+                comma = true;
 
             write(a);
         }
@@ -378,7 +379,7 @@ struct AstJsonEncoder : public AstVisitor
                 if (comma)
                     writeRaw(",");
                 else
-                    comma = false;
+                    comma = true;
                 write(prop);
             }
         });
@@ -612,6 +613,7 @@ struct AstJsonEncoder : public AstVisitor
         writeNode(node, "AstStatTypeAlias", [&]() {
             PROP(name);
             PROP(generics);
+            PROP(genericPacks);
             PROP(type);
             PROP(exported);
         });
@@ -664,13 +666,21 @@ struct AstJsonEncoder : public AstVisitor
         });
     }
 
+    void write(struct AstTypeOrPack node)
+    {
+        if (node.type)
+            write(node.type);
+        else
+            write(node.typePack);
+    }
+
     void write(class AstTypeReference* node)
     {
         writeNode(node, "AstTypeReference", [&]() {
             if (node->hasPrefix)
                 PROP(prefix);
             PROP(name);
-            PROP(generics);
+            PROP(parameters);
         });
     }
 
@@ -734,6 +744,13 @@ struct AstJsonEncoder : public AstVisitor
         });
     }
 
+    void write(class AstTypePackExplicit* node)
+    {
+        writeNode(node, "AstTypePackExplicit", [&]() {
+            PROP(typeList);
+        });
+    }
+
     void write(class AstTypePackVariadic* node)
     {
         writeNode(node, "AstTypePackVariadic", [&]() {
@@ -1018,6 +1035,12 @@ struct AstJsonEncoder : public AstVisitor
         return false;
     }
 
+    bool visit(class AstTypePackExplicit* node) override
+    {
+        write(node);
+        return false;
+    }
+
     bool visit(class AstTypePackVariadic* node) override
     {
         write(node);
