@@ -1,6 +1,7 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
 #include "Luau/ToString.h"
 
+#include "Luau/Scope.h"
 #include "Luau/TypeInfer.h"
 #include "Luau/TypePack.h"
 #include "Luau/TypeVar.h"
@@ -9,10 +10,15 @@
 #include <algorithm>
 #include <stdexcept>
 
-LUAU_FASTFLAG(LuauToStringFollowsBoundTo)
-LUAU_FASTFLAG(LuauExtraNilRecovery)
 LUAU_FASTFLAG(LuauOccursCheckOkWithRecursiveFunctions)
-LUAU_FASTFLAGVARIABLE(LuauInstantiatedTypeParamRecursion, false)
+LUAU_FASTFLAGVARIABLE(LuauFunctionArgumentNameSize, false)
+
+/*
+ * Prefix generic typenames with gen-
+ * Additionally, free types will be prefixed with free- and suffixed with their level.  eg free-a-4
+ * Fair warning: Setting this will break a lot of Luau unit tests.
+ */
+LUAU_FASTFLAGVARIABLE(DebugLuauVerboseTypeNames, false)
 
 namespace Luau
 {
@@ -59,6 +65,10 @@ struct FindCyclicTypes
         {
             for (TypeId itp : ttv.instantiatedTypeParams)
                 visitTypeVar(itp, *this, seen);
+
+            for (TypePackId itp : ttv.instantiatedTypePackParams)
+                visitTypeVar(itp, *this, seen);
+
             return exhaustive;
         }
 
@@ -151,15 +161,6 @@ struct StringifierState
             seen.erase(iter);
     }
 
-    static std::string generateName(size_t i)
-    {
-        std::string n;
-        n = char('a' + i % 26);
-        if (i >= 26)
-            n += std::to_string(i / 26);
-        return n;
-    }
-
     std::string getName(TypeId ty)
     {
         const size_t s = result.nameMap.typeVars.size();
@@ -239,15 +240,6 @@ struct TypeVarStringifier
             return;
         }
 
-        if (!FFlag::LuauAddMissingFollow)
-        {
-            if (get<FreeTypeVar>(tv))
-            {
-                state.emit(state.getName(tv));
-                return;
-            }
-        }
-
         Luau::visit(
             [this, tv](auto&& t) {
                 return (*this)(tv, t);
@@ -258,34 +250,62 @@ struct TypeVarStringifier
     void stringify(TypePackId tp);
     void stringify(TypePackId tpid, const std::vector<std::optional<FunctionArgument>>& names);
 
-    void stringify(const std::vector<TypeId>& types)
+    void stringify(const std::vector<TypeId>& types, const std::vector<TypePackId>& typePacks)
     {
-        if (types.size() == 0)
+        if (types.size() == 0 && typePacks.size() == 0)
             return;
 
-        if (types.size())
+        if (types.size() || typePacks.size())
             state.emit("<");
 
-        for (size_t i = 0; i < types.size(); ++i)
-        {
-            if (i > 0)
-                state.emit(", ");
+        bool first = true;
 
-            stringify(types[i]);
+        for (TypeId ty : types)
+        {
+            if (!first)
+                state.emit(", ");
+            first = false;
+
+            stringify(ty);
         }
 
-        if (types.size())
+        bool singleTp = typePacks.size() == 1;
+
+        for (TypePackId tp : typePacks)
+        {
+            if (isEmpty(tp) && singleTp)
+                continue;
+
+            if (!first)
+                state.emit(", ");
+            else
+                first = false;
+
+            if (!singleTp)
+                state.emit("(");
+
+            stringify(tp);
+
+            if (!singleTp)
+                state.emit(")");
+        }
+
+        if (types.size() || typePacks.size())
             state.emit(">");
     }
 
     void operator()(TypeId ty, const Unifiable::Free& ftv)
     {
         state.result.invalid = true;
+        if (FFlag::DebugLuauVerboseTypeNames)
+            state.emit("free-");
+        state.emit(state.getName(ty));
 
-        if (FFlag::LuauAddMissingFollow)
-            state.emit(state.getName(ty));
-        else
-            state.emit("<FREE>");
+        if (FFlag::DebugLuauVerboseTypeNames)
+        {
+            state.emit("-");
+            state.emit(std::to_string(ftv.level.level));
+        }
     }
 
     void operator()(TypeId, const BoundTypeVar& btv)
@@ -329,6 +349,23 @@ struct TypeVarStringifier
         }
     }
 
+    void operator()(TypeId, const SingletonTypeVar& stv)
+    {
+        if (const BoolSingleton* bs = Luau::get<BoolSingleton>(&stv))
+            state.emit(bs->value ? "true" : "false");
+        else if (const StringSingleton* ss = Luau::get<StringSingleton>(&stv))
+        {
+            state.emit("\"");
+            state.emit(escape(ss->value));
+            state.emit("\"");
+        }
+        else
+        {
+            LUAU_ASSERT(!"Unknown singleton type");
+            throw std::runtime_error("Unknown singleton type");
+        }
+    }
+
     void operator()(TypeId, const FunctionTypeVar& ftv)
     {
         if (state.hasSeen(&ftv))
@@ -338,6 +375,7 @@ struct TypeVarStringifier
             return;
         }
 
+        // We should not be respecting opts.hideNamedFunctionTypeParameters here.
         if (ftv.generics.size() > 0 || ftv.genericPacks.size() > 0)
         {
             state.emit("<");
@@ -388,7 +426,7 @@ struct TypeVarStringifier
 
     void operator()(TypeId, const TableTypeVar& ttv)
     {
-        if (FFlag::LuauToStringFollowsBoundTo && ttv.boundTo)
+        if (ttv.boundTo)
             return stringify(*ttv.boundTo);
 
         if (!state.exhaustive)
@@ -411,14 +449,14 @@ struct TypeVarStringifier
                 }
 
                 state.emit(*ttv.name);
-                stringify(ttv.instantiatedTypeParams);
+                stringify(ttv.instantiatedTypeParams, ttv.instantiatedTypePackParams);
                 return;
             }
             if (ttv.syntheticName)
             {
                 state.result.invalid = true;
                 state.emit(*ttv.syntheticName);
-                stringify(ttv.instantiatedTypeParams);
+                stringify(ttv.instantiatedTypeParams, ttv.instantiatedTypePackParams);
                 return;
             }
         }
@@ -493,7 +531,14 @@ struct TypeVarStringifier
                 break;
             }
 
-            state.emit(name);
+            if (isIdentifier(name))
+                state.emit(name);
+            else
+            {
+                state.emit("[\"");
+                state.emit(escape(name));
+                state.emit("\"]");
+            }
             state.emit(": ");
             stringify(prop.type);
             comma = true;
@@ -539,8 +584,7 @@ struct TypeVarStringifier
         std::vector<std::string> results = {};
         for (auto el : &uv)
         {
-            if (FFlag::LuauExtraNilRecovery || FFlag::LuauAddMissingFollow)
-                el = follow(el);
+            el = follow(el);
 
             if (isNil(el))
             {
@@ -604,8 +648,7 @@ struct TypeVarStringifier
         std::vector<std::string> results = {};
         for (auto el : uv.parts)
         {
-            if (FFlag::LuauExtraNilRecovery || FFlag::LuauAddMissingFollow)
-                el = follow(el);
+            el = follow(el);
 
             std::string saved = std::move(state.result.name);
 
@@ -691,16 +734,6 @@ struct TypePackStringifier
             return;
         }
 
-        if (!FFlag::LuauAddMissingFollow)
-        {
-            if (get<FreeTypePack>(tp))
-            {
-                state.emit(state.getName(tp));
-                state.emit("...");
-                return;
-            }
-        }
-
         auto it = state.cycleTpNames.find(tp);
         if (it != state.cycleTpNames.end())
         {
@@ -733,12 +766,23 @@ struct TypePackStringifier
             else
                 state.emit(", ");
 
-            LUAU_ASSERT(elemNames.empty() || elemIndex < elemNames.size());
-
-            if (!elemNames.empty() && elemNames[elemIndex])
+            if (FFlag::LuauFunctionArgumentNameSize)
             {
-                state.emit(elemNames[elemIndex]->name);
-                state.emit(": ");
+                if (elemIndex < elemNames.size() && elemNames[elemIndex])
+                {
+                    state.emit(elemNames[elemIndex]->name);
+                    state.emit(": ");
+                }
+            }
+            else
+            {
+                LUAU_ASSERT(elemNames.empty() || elemIndex < elemNames.size());
+
+                if (!elemNames.empty() && elemNames[elemIndex])
+                {
+                    state.emit(elemNames[elemIndex]->name);
+                    state.emit(": ");
+                }
             }
             elemIndex++;
 
@@ -773,6 +817,8 @@ struct TypePackStringifier
 
     void operator()(TypePackId tp, const GenericTypePack& pack)
     {
+        if (FFlag::DebugLuauVerboseTypeNames)
+            state.emit("gen-");
         if (pack.explicitName)
         {
             state.result.nameMap.typePacks[tp] = pack.name;
@@ -788,16 +834,17 @@ struct TypePackStringifier
     void operator()(TypePackId tp, const FreeTypePack& pack)
     {
         state.result.invalid = true;
+        if (FFlag::DebugLuauVerboseTypeNames)
+            state.emit("free-");
+        state.emit(state.getName(tp));
 
-        if (FFlag::LuauAddMissingFollow)
+        if (FFlag::DebugLuauVerboseTypeNames)
         {
-            state.emit(state.getName(tp));
-            state.emit("...");
-        }
-        else
-        {
-            state.emit("<FREETP>");
+            state.emit("-");
+            state.emit(std::to_string(pack.level.level));
         }
+
+        state.emit("...");
     }
 
     void operator()(TypePackId, const BoundTypePack& btv)
@@ -831,23 +878,15 @@ static void assignCycleNames(const std::unordered_set<TypeId>& cycles, const std
         std::string name;
 
         // TODO: use the stringified type list if there are no cycles
-        if (FFlag::LuauInstantiatedTypeParamRecursion)
+        if (auto ttv = get<TableTypeVar>(follow(cycleTy)); !exhaustive && ttv && (ttv->syntheticName || ttv->name))
         {
-            if (auto ttv = get<TableTypeVar>(follow(cycleTy)); !exhaustive && ttv && (ttv->syntheticName || ttv->name))
-            {
-                // If we have a cycle type in type parameters, assign a cycle name for this named table
-                if (std::find_if(ttv->instantiatedTypeParams.begin(), ttv->instantiatedTypeParams.end(), [&](auto&& el) {
-                        return cycles.count(follow(el));
-                    }) != ttv->instantiatedTypeParams.end())
-                    cycleNames[cycleTy] = ttv->name ? *ttv->name : *ttv->syntheticName;
+            // If we have a cycle type in type parameters, assign a cycle name for this named table
+            if (std::find_if(ttv->instantiatedTypeParams.begin(), ttv->instantiatedTypeParams.end(), [&](auto&& el) {
+                    return cycles.count(follow(el));
+                }) != ttv->instantiatedTypeParams.end())
+                cycleNames[cycleTy] = ttv->name ? *ttv->name : *ttv->syntheticName;
 
-                continue;
-            }
-        }
-        else
-        {
-            if (auto ttv = get<TableTypeVar>(follow(cycleTy)); !exhaustive && ttv && (ttv->syntheticName || ttv->name))
-                continue;
+            continue;
         }
 
         name = "t" + std::to_string(nextIndex);
@@ -879,45 +918,6 @@ ToStringResult toStringDetailed(TypeId ty, const ToStringOptions& opts)
 
     ToStringResult result;
 
-    if (!FFlag::LuauInstantiatedTypeParamRecursion && !opts.exhaustive)
-    {
-        if (auto ttv = get<TableTypeVar>(ty); ttv && (ttv->name || ttv->syntheticName))
-        {
-            if (ttv->syntheticName)
-                result.invalid = true;
-
-            // If scope if provided, add module name and check visibility
-            if (ttv->name && opts.scope)
-            {
-                auto [success, moduleName] = canUseTypeNameInScope(opts.scope, *ttv->name);
-
-                if (!success)
-                    result.invalid = true;
-
-                if (moduleName)
-                    result.name = format("%s.", moduleName->c_str());
-            }
-
-            result.name += ttv->name ? *ttv->name : *ttv->syntheticName;
-
-            if (ttv->instantiatedTypeParams.empty())
-                return result;
-
-            std::vector<std::string> params;
-            for (TypeId tp : ttv->instantiatedTypeParams)
-                params.push_back(toString(tp));
-
-            result.name += "<" + join(params, ", ") + ">";
-            return result;
-        }
-        else if (auto mtv = get<MetatableTypeVar>(ty); mtv && mtv->syntheticName)
-        {
-            result.invalid = true;
-            result.name = *mtv->syntheticName;
-            return result;
-        }
-    }
-
     StringifierState state{opts, result, opts.nameMap};
 
     std::unordered_set<TypeId> cycles;
@@ -929,7 +929,7 @@ ToStringResult toStringDetailed(TypeId ty, const ToStringOptions& opts)
 
     TypeVarStringifier tvs{state};
 
-    if (FFlag::LuauInstantiatedTypeParamRecursion && !opts.exhaustive)
+    if (!opts.exhaustive)
     {
         if (auto ttv = get<TableTypeVar>(ty); ttv && (ttv->name || ttv->syntheticName))
         {
@@ -950,31 +950,7 @@ ToStringResult toStringDetailed(TypeId ty, const ToStringOptions& opts)
 
             result.name += ttv->name ? *ttv->name : *ttv->syntheticName;
 
-            if (ttv->instantiatedTypeParams.empty())
-                return result;
-
-            result.name += "<";
-
-            bool first = true;
-            for (TypeId ty : ttv->instantiatedTypeParams)
-            {
-                if (!first)
-                    result.name += ", ";
-                else
-                    first = false;
-
-                tvs.stringify(ty);
-            }
-
-            if (opts.maxTypeLength > 0 && result.name.length() > opts.maxTypeLength)
-            {
-                result.truncated = true;
-                result.name += "... <TRUNCATED>";
-            }
-            else
-            {
-                result.name += ">";
-            }
+            tvs.stringify(ttv->instantiatedTypeParams, ttv->instantiatedTypePackParams);
 
             return result;
         }
@@ -1123,20 +1099,141 @@ std::string toString(const TypePackVar& tp, const ToStringOptions& opts)
     return toString(const_cast<TypePackId>(&tp), std::move(opts));
 }
 
-void dump(TypeId ty)
+std::string toStringNamedFunction(const std::string& prefix, const FunctionTypeVar& ftv, ToStringOptions opts)
 {
-    ToStringOptions opts;
-    opts.exhaustive = true;
-    opts.functionTypeArguments = true;
-    printf("%s\n", toString(ty, opts).c_str());
+    std::string s = prefix;
+
+    auto toString_ = [&opts](TypeId ty) -> std::string {
+        ToStringResult res = toStringDetailed(ty, opts);
+        opts.nameMap = std::move(res.nameMap);
+        return res.name;
+    };
+
+    auto toStringPack_ = [&opts](TypePackId ty) -> std::string {
+        ToStringResult res = toStringDetailed(ty, opts);
+        opts.nameMap = std::move(res.nameMap);
+        return res.name;
+    };
+
+    if (!opts.hideNamedFunctionTypeParameters && (!ftv.generics.empty() || !ftv.genericPacks.empty()))
+    {
+        s += "<";
+
+        bool first = true;
+        for (TypeId g : ftv.generics)
+        {
+            if (!first)
+                s += ", ";
+            first = false;
+            s += toString_(g);
+        }
+
+        for (TypePackId gp : ftv.genericPacks)
+        {
+            if (!first)
+                s += ", ";
+            first = false;
+            s += toStringPack_(gp);
+        }
+
+        s += ">";
+    }
+
+    s += "(";
+
+    auto argPackIter = begin(ftv.argTypes);
+    auto argNameIter = ftv.argNames.begin();
+
+    bool first = true;
+    while (argPackIter != end(ftv.argTypes))
+    {
+        if (!first)
+            s += ", ";
+        first = false;
+
+        if (FFlag::LuauFunctionArgumentNameSize)
+        {
+            // We don't currently respect opts.functionTypeArguments. I don't think this function should.
+            if (argNameIter != ftv.argNames.end())
+            {
+                s += (*argNameIter ? (*argNameIter)->name : "_") + ": ";
+                ++argNameIter;
+            }
+            else
+            {
+                s += "_: ";
+            }
+        }
+        else
+        {
+            // argNames is guaranteed to be equal to argTypes iff argNames is not empty.
+            // We don't currently respect opts.functionTypeArguments. I don't think this function should.
+            if (!ftv.argNames.empty())
+                s += (*argNameIter ? (*argNameIter)->name : "_") + ": ";
+        }
+
+        s += toString_(*argPackIter);
+        ++argPackIter;
+
+        if (!FFlag::LuauFunctionArgumentNameSize)
+        {
+            if (!ftv.argNames.empty())
+            {
+                LUAU_ASSERT(argNameIter != ftv.argNames.end());
+                ++argNameIter;
+            }
+        }
+    }
+
+    if (argPackIter.tail())
+    {
+        if (auto vtp = get<VariadicTypePack>(*argPackIter.tail()))
+            s += ", ...: " + toString_(vtp->ty);
+        else
+            s += ", ...: " + toStringPack_(*argPackIter.tail());
+    }
+
+    s += "): ";
+
+    size_t retSize = size(ftv.retType);
+    bool hasTail = !finite(ftv.retType);
+    if (retSize == 0 && !hasTail)
+        s += "()";
+    else if ((retSize == 0 && hasTail) || (retSize == 1 && !hasTail))
+        s += toStringPack_(ftv.retType);
+    else
+        s += "(" + toStringPack_(ftv.retType) + ")";
+
+    return s;
 }
 
-void dump(TypePackId ty)
+std::string dump(TypeId ty)
 {
     ToStringOptions opts;
     opts.exhaustive = true;
     opts.functionTypeArguments = true;
-    printf("%s\n", toString(ty, opts).c_str());
+    std::string s = toString(ty, opts);
+    printf("%s\n", s.c_str());
+    return s;
+}
+
+std::string dump(TypePackId ty)
+{
+    ToStringOptions opts;
+    opts.exhaustive = true;
+    opts.functionTypeArguments = true;
+    std::string s = toString(ty, opts);
+    printf("%s\n", s.c_str());
+    return s;
+}
+
+std::string generateName(size_t i)
+{
+    std::string n;
+    n = char('a' + i % 26);
+    if (i >= 26)
+        n += std::to_string(i / 26);
+    return n;
 }
 
 } // namespace Luau
