@@ -255,6 +255,14 @@ public:
     {
         return visit((class AstType*)node);
     }
+    virtual bool visit(class AstTypeSingletonBool* node)
+    {
+        return visit((class AstType*)node);
+    }
+    virtual bool visit(class AstTypeSingletonString* node)
+    {
+        return visit((class AstType*)node);
+    }
     virtual bool visit(class AstTypeError* node)
     {
         return visit((class AstType*)node);
@@ -264,6 +272,10 @@ public:
     {
         return false;
     }
+    virtual bool visit(class AstTypePackExplicit* node)
+    {
+        return visit((class AstTypePack*)node);
+    }
     virtual bool visit(class AstTypePackVariadic* node)
     {
         return visit((class AstTypePack*)node);
@@ -930,12 +942,14 @@ class AstStatTypeAlias : public AstStat
 public:
     LUAU_RTTI(AstStatTypeAlias)
 
-    AstStatTypeAlias(const Location& location, const AstName& name, const AstArray<AstName>& generics, AstType* type, bool exported);
+    AstStatTypeAlias(const Location& location, const AstName& name, const AstArray<AstName>& generics, const AstArray<AstName>& genericPacks,
+        AstType* type, bool exported);
 
     void visit(AstVisitor* visitor) override;
 
     AstName name;
     AstArray<AstName> generics;
+    AstArray<AstName> genericPacks;
     AstType* type;
     bool exported;
 };
@@ -1007,19 +1021,28 @@ public:
     }
 };
 
+// Don't have Luau::Variant available, it's a bit of an overhead, but a plain struct is nice to use
+struct AstTypeOrPack
+{
+    AstType* type = nullptr;
+    AstTypePack* typePack = nullptr;
+};
+
 class AstTypeReference : public AstType
 {
 public:
     LUAU_RTTI(AstTypeReference)
 
-    AstTypeReference(const Location& location, std::optional<AstName> prefix, AstName name, const AstArray<AstType*>& generics = {});
+    AstTypeReference(const Location& location, std::optional<AstName> prefix, AstName name, bool hasParameterList = false,
+        const AstArray<AstTypeOrPack>& parameters = {});
 
     void visit(AstVisitor* visitor) override;
 
     bool hasPrefix;
+    bool hasParameterList;
     AstName prefix;
     AstName name;
-    AstArray<AstType*> generics;
+    AstArray<AstTypeOrPack> parameters;
 };
 
 struct AstTableProp
@@ -1143,6 +1166,30 @@ public:
     unsigned messageIndex;
 };
 
+class AstTypeSingletonBool : public AstType
+{
+public:
+    LUAU_RTTI(AstTypeSingletonBool)
+
+    AstTypeSingletonBool(const Location& location, bool value);
+
+    void visit(AstVisitor* visitor) override;
+
+    bool value;
+};
+
+class AstTypeSingletonString : public AstType
+{
+public:
+    LUAU_RTTI(AstTypeSingletonString)
+
+    AstTypeSingletonString(const Location& location, const AstArray<char>& value);
+
+    void visit(AstVisitor* visitor) override;
+
+    const AstArray<char> value;
+};
+
 class AstTypePack : public AstNode
 {
 public:
@@ -1152,6 +1199,18 @@ public:
     }
 };
 
+class AstTypePackExplicit : public AstTypePack
+{
+public:
+    LUAU_RTTI(AstTypePackExplicit)
+
+    AstTypePackExplicit(const Location& location, AstTypeList typeList);
+
+    void visit(AstVisitor* visitor) override;
+
+    AstTypeList typeList;
+};
+
 class AstTypePackVariadic : public AstTypePack
 {
 public:
