@@ -8,12 +8,13 @@
 #include "lfunc.h"
 #include "lgc.h"
 #include "ldo.h"
+#include "ludata.h"
 #include "lvm.h"
 #include "lnumutils.h"
 
 #include <string.h>
 
-LUAU_FASTFLAG(LuauGcFullSkipInactiveThreads)
+LUAU_FASTFLAG(LuauActivateBeforeExec)
 
 const char* lua_ident = "$Lua: Lua 5.1.4 Copyright (C) 1994-2008 Lua.org, PUC-Rio $\n"
                         "$Authors: R. Ierusalimschy, L. H. de Figueiredo & W. Celes $\n"
@@ -43,36 +44,30 @@ static Table* getcurrenv(lua_State* L)
     }
 }
 
-static LUAU_NOINLINE TValue* index2adrslow(lua_State* L, int idx)
+static LUAU_NOINLINE TValue* pseudo2addr(lua_State* L, int idx)
 {
-    api_check(L, idx <= 0);
-    if (idx > LUA_REGISTRYINDEX)
+    api_check(L, lua_ispseudo(idx));
+    switch (idx)
+    { /* pseudo-indices */
+    case LUA_REGISTRYINDEX:
+        return registry(L);
+    case LUA_ENVIRONINDEX:
     {
-        api_check(L, idx != 0 && -idx <= L->top - L->base);
-        return L->top + idx;
+        sethvalue(L, &L->env, getcurrenv(L));
+        return &L->env;
+    }
+    case LUA_GLOBALSINDEX:
+        return gt(L);
+    default:
+    {
+        Closure* func = curr_func(L);
+        idx = LUA_GLOBALSINDEX - idx;
+        return (idx <= func->nupvalues) ? &func->c.upvals[idx - 1] : cast_to(TValue*, luaO_nilobject);
+    }
     }
-    else
-        switch (idx)
-        { /* pseudo-indices */
-        case LUA_REGISTRYINDEX:
-            return registry(L);
-        case LUA_ENVIRONINDEX:
-        {
-            sethvalue(L, &L->env, getcurrenv(L));
-            return &L->env;
-        }
-        case LUA_GLOBALSINDEX:
-            return gt(L);
-        default:
-        {
-            Closure* func = curr_func(L);
-            idx = LUA_GLOBALSINDEX - idx;
-            return (idx <= func->nupvalues) ? &func->c.upvals[idx - 1] : cast_to(TValue*, luaO_nilobject);
-        }
-        }
 }
 
-static LUAU_FORCEINLINE TValue* index2adr(lua_State* L, int idx)
+static LUAU_FORCEINLINE TValue* index2addr(lua_State* L, int idx)
 {
     if (idx > 0)
     {
@@ -83,15 +78,20 @@ static LUAU_FORCEINLINE TValue* index2adr(lua_State* L, int idx)
         else
             return o;
     }
+    else if (idx > LUA_REGISTRYINDEX)
+    {
+        api_check(L, idx != 0 && -idx <= L->top - L->base);
+        return L->top + idx;
+    }
     else
     {
-        return index2adrslow(L, idx);
+        return pseudo2addr(L, idx);
     }
 }
 
 const TValue* luaA_toobject(lua_State* L, int idx)
 {
-    StkId p = index2adr(L, idx);
+    StkId p = index2addr(L, idx);
     return (p == luaO_nilobject) ? NULL : p;
 }
 
@@ -145,7 +145,7 @@ void lua_xpush(lua_State* from, lua_State* to, int idx)
 {
     api_check(from, from->global == to->global);
     luaC_checkthreadsleep(to);
-    setobj2s(to, to->top, index2adr(from, idx));
+    setobj2s(to, to->top, index2addr(from, idx));
     api_incr_top(to);
     return;
 }
@@ -172,6 +172,12 @@ lua_State* lua_mainthread(lua_State* L)
 ** basic stack manipulation
 */
 
+int lua_absindex(lua_State* L, int idx)
+{
+    api_check(L, (idx > 0 && idx <= L->top - L->base) || (idx < 0 && -idx <= L->top - L->base) || lua_ispseudo(idx));
+    return idx > 0 || lua_ispseudo(idx) ? idx : cast_int(L->top - L->base) + idx + 1;
+}
+
 int lua_gettop(lua_State* L)
 {
     return cast_int(L->top - L->base);
@@ -196,7 +202,7 @@ void lua_settop(lua_State* L, int idx)
 
 void lua_remove(lua_State* L, int idx)
 {
-    StkId p = index2adr(L, idx);
+    StkId p = index2addr(L, idx);
     api_checkvalidindex(L, p);
     while (++p < L->top)
         setobjs2s(L, p - 1, p);
@@ -207,7 +213,7 @@ void lua_remove(lua_State* L, int idx)
 void lua_insert(lua_State* L, int idx)
 {
     luaC_checkthreadsleep(L);
-    StkId p = index2adr(L, idx);
+    StkId p = index2addr(L, idx);
     api_checkvalidindex(L, p);
     for (StkId q = L->top; q > p; q--)
         setobjs2s(L, q, q - 1);
@@ -222,7 +228,7 @@ void lua_replace(lua_State* L, int idx)
         luaG_runerror(L, "no calling environment");
     api_checknelems(L, 1);
     luaC_checkthreadsleep(L);
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     api_checkvalidindex(L, o);
     if (idx == LUA_ENVIRONINDEX)
     {
@@ -244,7 +250,7 @@ void lua_replace(lua_State* L, int idx)
 void lua_pushvalue(lua_State* L, int idx)
 {
     luaC_checkthreadsleep(L);
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     setobj2s(L, L->top, o);
     api_incr_top(L);
     return;
@@ -256,7 +262,7 @@ void lua_pushvalue(lua_State* L, int idx)
 
 int lua_type(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return (o == luaO_nilobject) ? LUA_TNONE : ttype(o);
 }
 
@@ -267,20 +273,20 @@ const char* lua_typename(lua_State* L, int t)
 
 int lua_iscfunction(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return iscfunction(o);
 }
 
 int lua_isLfunction(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return isLfunction(o);
 }
 
 int lua_isnumber(lua_State* L, int idx)
 {
     TValue n;
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     return tonumber(o, &n);
 }
 
@@ -292,14 +298,14 @@ int lua_isstring(lua_State* L, int idx)
 
 int lua_isuserdata(lua_State* L, int idx)
 {
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     return (ttisuserdata(o) || ttislightuserdata(o));
 }
 
 int lua_rawequal(lua_State* L, int index1, int index2)
 {
-    StkId o1 = index2adr(L, index1);
-    StkId o2 = index2adr(L, index2);
+    StkId o1 = index2addr(L, index1);
+    StkId o2 = index2addr(L, index2);
     return (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0 : luaO_rawequalObj(o1, o2);
 }
 
@@ -307,8 +313,8 @@ int lua_equal(lua_State* L, int index1, int index2)
 {
     StkId o1, o2;
     int i;
-    o1 = index2adr(L, index1);
-    o2 = index2adr(L, index2);
+    o1 = index2addr(L, index1);
+    o2 = index2addr(L, index2);
     i = (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0 : equalobj(L, o1, o2);
     return i;
 }
@@ -317,8 +323,8 @@ int lua_lessthan(lua_State* L, int index1, int index2)
 {
     StkId o1, o2;
     int i;
-    o1 = index2adr(L, index1);
-    o2 = index2adr(L, index2);
+    o1 = index2addr(L, index1);
+    o2 = index2addr(L, index2);
     i = (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0 : luaV_lessthan(L, o1, o2);
     return i;
 }
@@ -326,7 +332,7 @@ int lua_lessthan(lua_State* L, int index1, int index2)
 double lua_tonumberx(lua_State* L, int idx, int* isnum)
 {
     TValue n;
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     if (tonumber(o, &n))
     {
         if (isnum)
@@ -344,7 +350,7 @@ double lua_tonumberx(lua_State* L, int idx, int* isnum)
 int lua_tointegerx(lua_State* L, int idx, int* isnum)
 {
     TValue n;
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     if (tonumber(o, &n))
     {
         int res;
@@ -365,7 +371,7 @@ int lua_tointegerx(lua_State* L, int idx, int* isnum)
 unsigned lua_tounsignedx(lua_State* L, int idx, int* isnum)
 {
     TValue n;
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     if (tonumber(o, &n))
     {
         unsigned res;
@@ -385,13 +391,13 @@ unsigned lua_tounsignedx(lua_State* L, int idx, int* isnum)
 
 int lua_toboolean(lua_State* L, int idx)
 {
-    const TValue* o = index2adr(L, idx);
+    const TValue* o = index2addr(L, idx);
     return !l_isfalse(o);
 }
 
 const char* lua_tolstring(lua_State* L, int idx, size_t* len)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     if (!ttisstring(o))
     {
         luaC_checkthreadsleep(L);
@@ -402,7 +408,7 @@ const char* lua_tolstring(lua_State* L, int idx, size_t* len)
             return NULL;
         }
         luaC_checkGC(L);
-        o = index2adr(L, idx); /* previous call may reallocate the stack */
+        o = index2addr(L, idx); /* previous call may reallocate the stack */
     }
     if (len != NULL)
         *len = tsvalue(o)->len;
@@ -411,7 +417,7 @@ const char* lua_tolstring(lua_State* L, int idx, size_t* len)
 
 const char* lua_tostringatom(lua_State* L, int idx, int* atom)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     if (!ttisstring(o))
         return NULL;
     const TString* s = tsvalue(o);
@@ -432,7 +438,7 @@ const char* lua_namecallatom(lua_State* L, int* atom)
 
 const float* lua_tovector(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     if (!ttisvector(o))
     {
         return NULL;
@@ -442,7 +448,7 @@ const float* lua_tovector(lua_State* L, int idx)
 
 int lua_objlen(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     switch (ttype(o))
     {
     case LUA_TSTRING:
@@ -463,13 +469,13 @@ int lua_objlen(lua_State* L, int idx)
 
 lua_CFunction lua_tocfunction(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return (!iscfunction(o)) ? NULL : cast_to(lua_CFunction, clvalue(o)->c.f);
 }
 
 void* lua_touserdata(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     switch (ttype(o))
     {
     case LUA_TUSERDATA:
@@ -483,13 +489,13 @@ void* lua_touserdata(lua_State* L, int idx)
 
 void* lua_touserdatatagged(lua_State* L, int idx, int tag)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return (ttisuserdata(o) && uvalue(o)->tag == tag) ? uvalue(o)->data : NULL;
 }
 
 int lua_userdatatag(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     if (ttisuserdata(o))
         return uvalue(o)->tag;
     return -1;
@@ -497,13 +503,13 @@ int lua_userdatatag(lua_State* L, int idx)
 
 lua_State* lua_tothread(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     return (!ttisthread(o)) ? NULL : thvalue(o);
 }
 
 const void* lua_topointer(lua_State* L, int idx)
 {
-    StkId o = index2adr(L, idx);
+    StkId o = index2addr(L, idx);
     switch (ttype(o))
     {
     case LUA_TTABLE:
@@ -552,12 +558,21 @@ void lua_pushunsigned(lua_State* L, unsigned u)
     return;
 }
 
-void lua_pushvector(lua_State* L, float x, float y, float z)
+#if LUA_VECTOR_SIZE == 4
+void lua_pushvector(lua_State* L, float x, float y, float z, float w)
 {
-    setvvalue(L->top, x, y, z);
+    setvvalue(L->top, x, y, z, w);
     api_incr_top(L);
     return;
 }
+#else
+void lua_pushvector(lua_State* L, float x, float y, float z)
+{
+    setvvalue(L->top, x, y, z, 0.0f);
+    api_incr_top(L);
+    return;
+}
+#endif
 
 void lua_pushlstring(lua_State* L, const char* s, size_t len)
 {
@@ -595,7 +610,7 @@ const char* lua_pushfstringL(lua_State* L, const char* fmt, ...)
     return ret;
 }
 
-void lua_pushcfunction(lua_State* L, lua_CFunction fn, const char* debugname, int nup, lua_Continuation cont)
+void lua_pushcclosurek(lua_State* L, lua_CFunction fn, const char* debugname, int nup, lua_Continuation cont)
 {
     luaC_checkGC(L);
     luaC_checkthreadsleep(L);
@@ -642,7 +657,7 @@ int lua_pushthread(lua_State* L)
 void lua_gettable(lua_State* L, int idx)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_checkvalidindex(L, t);
     luaV_gettable(L, t, L->top - 1, L->top - 1);
     return;
@@ -651,7 +666,7 @@ void lua_gettable(lua_State* L, int idx)
 void lua_getfield(lua_State* L, int idx, const char* k)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_checkvalidindex(L, t);
     TValue key;
     setsvalue(L, &key, luaS_new(L, k));
@@ -663,7 +678,7 @@ void lua_getfield(lua_State* L, int idx, const char* k)
 void lua_rawgetfield(lua_State* L, int idx, const char* k)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_check(L, ttistable(t));
     TValue key;
     setsvalue(L, &key, luaS_new(L, k));
@@ -675,7 +690,7 @@ void lua_rawgetfield(lua_State* L, int idx, const char* k)
 void lua_rawget(lua_State* L, int idx)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_check(L, ttistable(t));
     setobj2s(L, L->top - 1, luaH_get(hvalue(t), L->top - 1));
     return;
@@ -684,7 +699,7 @@ void lua_rawget(lua_State* L, int idx)
 void lua_rawgeti(lua_State* L, int idx, int n)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_check(L, ttistable(t));
     setobj2s(L, L->top, luaH_getnum(hvalue(t), n));
     api_incr_top(L);
@@ -700,30 +715,31 @@ void lua_createtable(lua_State* L, int narray, int nrec)
     return;
 }
 
-void lua_setreadonly(lua_State* L, int objindex, bool value)
+void lua_setreadonly(lua_State* L, int objindex, int enabled)
 {
-    const TValue* o = index2adr(L, objindex);
+    const TValue* o = index2addr(L, objindex);
     api_check(L, ttistable(o));
     Table* t = hvalue(o);
-    t->readonly = value;
+    api_check(L, t != hvalue(registry(L)));
+    t->readonly = bool(enabled);
     return;
 }
 
 int lua_getreadonly(lua_State* L, int objindex)
 {
-    const TValue* o = index2adr(L, objindex);
+    const TValue* o = index2addr(L, objindex);
     api_check(L, ttistable(o));
     Table* t = hvalue(o);
     int res = t->readonly;
     return res;
 }
 
-void lua_setsafeenv(lua_State* L, int objindex, bool value)
+void lua_setsafeenv(lua_State* L, int objindex, int enabled)
 {
-    const TValue* o = index2adr(L, objindex);
+    const TValue* o = index2addr(L, objindex);
     api_check(L, ttistable(o));
     Table* t = hvalue(o);
-    t->safeenv = value;
+    t->safeenv = bool(enabled);
     return;
 }
 
@@ -732,7 +748,7 @@ int lua_getmetatable(lua_State* L, int objindex)
     const TValue* obj;
     Table* mt = NULL;
     int res;
-    obj = index2adr(L, objindex);
+    obj = index2addr(L, objindex);
     switch (ttype(obj))
     {
     case LUA_TTABLE:
@@ -759,7 +775,7 @@ int lua_getmetatable(lua_State* L, int objindex)
 void lua_getfenv(lua_State* L, int idx)
 {
     StkId o;
-    o = index2adr(L, idx);
+    o = index2addr(L, idx);
     api_checkvalidindex(L, o);
     switch (ttype(o))
     {
@@ -785,7 +801,7 @@ void lua_settable(lua_State* L, int idx)
 {
     StkId t;
     api_checknelems(L, 2);
-    t = index2adr(L, idx);
+    t = index2addr(L, idx);
     api_checkvalidindex(L, t);
     luaV_settable(L, t, L->top - 2, L->top - 1);
     L->top -= 2; /* pop index and value */
@@ -797,7 +813,7 @@ void lua_setfield(lua_State* L, int idx, const char* k)
     StkId t;
     TValue key;
     api_checknelems(L, 1);
-    t = index2adr(L, idx);
+    t = index2addr(L, idx);
     api_checkvalidindex(L, t);
     setsvalue(L, &key, luaS_new(L, k));
     luaV_settable(L, t, &key, L->top - 1);
@@ -809,7 +825,7 @@ void lua_rawset(lua_State* L, int idx)
 {
     StkId t;
     api_checknelems(L, 2);
-    t = index2adr(L, idx);
+    t = index2addr(L, idx);
     api_check(L, ttistable(t));
     if (hvalue(t)->readonly)
         luaG_runerror(L, "Attempt to modify a readonly table");
@@ -823,7 +839,7 @@ void lua_rawseti(lua_State* L, int idx, int n)
 {
     StkId o;
     api_checknelems(L, 1);
-    o = index2adr(L, idx);
+    o = index2addr(L, idx);
     api_check(L, ttistable(o));
     if (hvalue(o)->readonly)
         luaG_runerror(L, "Attempt to modify a readonly table");
@@ -838,7 +854,7 @@ int lua_setmetatable(lua_State* L, int objindex)
     TValue* obj;
     Table* mt;
     api_checknelems(L, 1);
-    obj = index2adr(L, objindex);
+    obj = index2addr(L, objindex);
     api_checkvalidindex(L, obj);
     if (ttisnil(L->top - 1))
         mt = NULL;
@@ -880,7 +896,7 @@ int lua_setfenv(lua_State* L, int idx)
     StkId o;
     int res = 1;
     api_checknelems(L, 1);
-    o = index2adr(L, idx);
+    o = index2addr(L, idx);
     api_checkvalidindex(L, o);
     api_check(L, ttistable(L->top - 1));
     switch (ttype(o))
@@ -923,14 +939,21 @@ void lua_call(lua_State* L, int nargs, int nresults)
     checkresults(L, nargs, nresults);
     func = L->top - (nargs + 1);
 
-    int wasActive = luaC_threadactive(L);
-    l_setbit(L->stackstate, THREAD_ACTIVEBIT);
-    luaC_checkthreadsleep(L);
+    if (FFlag::LuauActivateBeforeExec)
+    {
+        luaD_call(L, func, nresults);
+    }
+    else
+    {
+        int oldactive = luaC_threadactive(L);
+        l_setbit(L->stackstate, THREAD_ACTIVEBIT);
+        luaC_checkthreadsleep(L);
 
-    luaD_call(L, func, nresults);
+        luaD_call(L, func, nresults);
 
-    if (!wasActive)
-        resetbit(L->stackstate, THREAD_ACTIVEBIT);
+        if (!oldactive)
+            resetbit(L->stackstate, THREAD_ACTIVEBIT);
+    }
 
     adjustresults(L, nresults);
     return;
@@ -964,21 +987,28 @@ int lua_pcall(lua_State* L, int nargs, int nresults, int errfunc)
         func = 0;
     else
     {
-        StkId o = index2adr(L, errfunc);
+        StkId o = index2addr(L, errfunc);
         api_checkvalidindex(L, o);
         func = savestack(L, o);
     }
     c.func = L->top - (nargs + 1); /* function to be called */
     c.nresults = nresults;
 
-    int wasActive = luaC_threadactive(L);
-    l_setbit(L->stackstate, THREAD_ACTIVEBIT);
-    luaC_checkthreadsleep(L);
+    if (FFlag::LuauActivateBeforeExec)
+    {
+        status = luaD_pcall(L, f_call, &c, savestack(L, c.func), func);
+    }
+    else
+    {
+        int oldactive = luaC_threadactive(L);
+        l_setbit(L->stackstate, THREAD_ACTIVEBIT);
+        luaC_checkthreadsleep(L);
 
-    status = luaD_pcall(L, f_call, &c, savestack(L, c.func), func);
+        status = luaD_pcall(L, f_call, &c, savestack(L, c.func), func);
 
-    if (!wasActive)
-        resetbit(L->stackstate, THREAD_ACTIVEBIT);
+        if (!oldactive)
+            resetbit(L->stackstate, THREAD_ACTIVEBIT);
+    }
 
     adjustresults(L, nresults);
     return status;
@@ -989,6 +1019,16 @@ int lua_status(lua_State* L)
     return L->status;
 }
 
+void* lua_getthreaddata(lua_State* L)
+{
+    return L->userdata;
+}
+
+void lua_setthreaddata(lua_State* L, void* data)
+{
+    L->userdata = data;
+}
+
 /*
 ** Garbage-collection function
 */
@@ -1021,6 +1061,11 @@ int lua_gc(lua_State* L, int what, int data)
         res = cast_int(g->totalbytes >> 10);
         break;
     }
+    case LUA_GCCOUNTB:
+    {
+        res = cast_int(g->totalbytes & 1023);
+        break;
+    }
     case LUA_GCISRUNNING:
     {
         res = (g->GCthreshold != SIZE_MAX);
@@ -1105,7 +1150,7 @@ l_noret lua_error(lua_State* L)
 int lua_next(lua_State* L, int idx)
 {
     luaC_checkthreadsleep(L);
-    StkId t = index2adr(L, idx);
+    StkId t = index2addr(L, idx);
     api_check(L, ttistable(t));
     int more = luaH_next(L, hvalue(t), L->top - 1);
     if (more)
@@ -1137,12 +1182,12 @@ void lua_concat(lua_State* L, int n)
     return;
 }
 
-void* lua_newuserdata(lua_State* L, size_t sz, int tag)
+void* lua_newuserdatatagged(lua_State* L, size_t sz, int tag)
 {
     api_check(L, unsigned(tag) < LUA_UTAG_LIMIT);
     luaC_checkGC(L);
     luaC_checkthreadsleep(L);
-    Udata* u = luaS_newudata(L, sz, tag);
+    Udata* u = luaU_newudata(L, sz, tag);
     setuvalue(L, L->top, u);
     api_incr_top(L);
     return u->data;
@@ -1152,8 +1197,8 @@ void* lua_newuserdatadtor(lua_State* L, size_t sz, void (*dtor)(void*))
 {
     luaC_checkGC(L);
     luaC_checkthreadsleep(L);
-    Udata* u = luaS_newudata(L, sz + sizeof(dtor), UTAG_IDTOR);
-    memcpy(u->data + sz, &dtor, sizeof(dtor));
+    Udata* u = luaU_newudata(L, sz + sizeof(dtor), UTAG_IDTOR);
+    memcpy(&u->data + sz, &dtor, sizeof(dtor));
     setuvalue(L, L->top, u);
     api_incr_top(L);
     return u->data;
@@ -1187,7 +1232,7 @@ const char* lua_getupvalue(lua_State* L, int funcindex, int n)
 {
     luaC_checkthreadsleep(L);
     TValue* val;
-    const char* name = aux_upvalue(index2adr(L, funcindex), n, &val);
+    const char* name = aux_upvalue(index2addr(L, funcindex), n, &val);
     if (name)
     {
         setobj2s(L, L->top, val);
@@ -1201,7 +1246,7 @@ const char* lua_setupvalue(lua_State* L, int funcindex, int n)
     const char* name;
     TValue* val;
     StkId fi;
-    fi = index2adr(L, funcindex);
+    fi = index2addr(L, funcindex);
     api_checknelems(L, 1);
     name = aux_upvalue(fi, n, &val);
     if (name)
@@ -1222,9 +1267,10 @@ uintptr_t lua_encodepointer(lua_State* L, uintptr_t p)
 
 int lua_ref(lua_State* L, int idx)
 {
+    api_check(L, idx != LUA_REGISTRYINDEX); /* idx is a stack index for value */
     int ref = LUA_REFNIL;
     global_State* g = L->global;
-    StkId p = index2adr(L, idx);
+    StkId p = index2addr(L, idx);
     if (!ttisnil(p))
     {
         Table* reg = hvalue(registry(L));
