@@ -8,11 +8,7 @@
 /* string size limit */
 #define MAXSSIZE (1 << 30)
 
-/* special tag value is used for user data with inline dtors */
-#define UTAG_IDTOR LUA_UTAG_LIMIT
-
 #define sizestring(len) (offsetof(TString, data) + len + 1)
-#define sizeudata(len) (offsetof(Udata, data) + len)
 
 #define luaS_new(L, s) (luaS_newlstr(L, s, strlen(s)))
 #define luaS_newliteral(L, s) (luaS_newlstr(L, "" s, (sizeof(s) / sizeof(char)) - 1))
@@ -26,8 +22,5 @@ LUAI_FUNC void luaS_resize(lua_State* L, int newsize);
 LUAI_FUNC TString* luaS_newlstr(lua_State* L, const char* str, size_t l);
 LUAI_FUNC void luaS_free(lua_State* L, TString* ts);
 
-LUAI_FUNC Udata* luaS_newudata(lua_State* L, size_t s, int tag);
-LUAI_FUNC void luaS_freeudata(lua_State* L, Udata* u);
-
 LUAI_FUNC TString* luaS_bufstart(lua_State* L, size_t size);
 LUAI_FUNC TString* luaS_buffinish(lua_State* L, TString* ts);
