@@ -10,10 +10,6 @@
 #include <sstream>
 #include <string_view>
 
-LUAU_FASTFLAG(LuauPreloadClosures)
-LUAU_FASTFLAG(LuauPreloadClosuresFenv)
-LUAU_FASTFLAG(LuauPreloadClosuresUpval)
-
 using namespace Luau;
 
 static std::string compileFunction(const char* source, uint32_t id)
@@ -75,20 +71,10 @@ TEST_CASE("BasicFunction")
     bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code);
     Luau::compileOrThrow(bcb, "local function foo(a, b) return b end");
 
-    if (FFlag::LuauPreloadClosures)
-    {
-        CHECK_EQ("\n" + bcb.dumpFunction(1), R"(
+    CHECK_EQ("\n" + bcb.dumpFunction(1), R"(
 DUPCLOSURE R0 K0
 RETURN R0 0
 )");
-    }
-    else
-    {
-        CHECK_EQ("\n" + bcb.dumpFunction(1), R"(
-NEWCLOSURE R0 P0
-RETURN R0 0
-)");
-    }
 
     CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
 RETURN R1 1
@@ -768,11 +754,11 @@ TEST_CASE("CaptureSelf")
 local MaterialsListClass = {}
 
 function MaterialsListClass:_MakeToolTip(guiElement, text)
-	local function updateTooltipPosition()
-		self._tweakingTooltipFrame = 5
-	end
+    local function updateTooltipPosition()
+        self._tweakingTooltipFrame = 5
+    end
 
-	updateTooltipPosition()
+    updateTooltipPosition()
 end
 
 return MaterialsListClass
@@ -1058,6 +1044,18 @@ RETURN R0 1
     CHECK_EQ("\n" + compileFunction0("return if false then 10 else 20"), R"(
 LOADN R0 20
 RETURN R0 1
+)");
+
+    // codegen for a true constant condition with non-constant expressions
+    CHECK_EQ("\n" + compileFunction0("return if true then {} else error()"), R"(
+NEWTABLE R0 0 0
+RETURN R0 1
+)");
+
+    // codegen for a false constant condition with non-constant expressions
+    CHECK_EQ("\n" + compileFunction0("return if false then error() else {}"), R"(
+NEWTABLE R0 0 0
+RETURN R0 1
 )");
 
     // codegen for a false (in this case 'nil') constant condition
@@ -1168,6 +1166,17 @@ RETURN R0 1
 )");
 }
 
+TEST_CASE("ConstantFoldStringLen")
+{
+    CHECK_EQ("\n" + compileFunction0("return #'string', #'', #'a', #('b')"), R"(
+LOADN R0 6
+LOADN R1 0
+LOADN R2 1
+LOADN R3 1
+RETURN R0 4
+)");
+}
+
 TEST_CASE("ConstantFoldCompare")
 {
     // ordered comparisons
@@ -2001,14 +2010,14 @@ TEST_CASE("UpvaluesLoopsBytecode")
 {
     CHECK_EQ("\n" + compileFunction(R"(
 function test()
-	for i=1,10 do
+    for i=1,10 do
         i = i
-		foo(function() return i end)
-		if bar then
-			break
-		end
-	end
-	return 0
+        foo(function() return i end)
+        if bar then
+            break
+        end
+    end
+    return 0
 end
 )",
                         1),
@@ -2035,14 +2044,14 @@ RETURN R0 1
 
     CHECK_EQ("\n" + compileFunction(R"(
 function test()
-	for i in ipairs(data) do
+    for i in ipairs(data) do
         i = i
-		foo(function() return i end)
-		if bar then
-			break
-		end
-	end
-	return 0
+        foo(function() return i end)
+        if bar then
+            break
+        end
+    end
+    return 0
 end
 )",
                         1),
@@ -2068,17 +2077,17 @@ RETURN R0 1
 
     CHECK_EQ("\n" + compileFunction(R"(
 function test()
-	local i = 0
-	while i < 5 do
-		local j
+    local i = 0
+    while i < 5 do
+        local j
         j = i
-		foo(function() return j end)
-		i = i + 1
-		if bar then
-			break
-		end
-	end
-	return 0
+        foo(function() return j end)
+        i = i + 1
+        if bar then
+            break
+        end
+    end
+    return 0
 end
 )",
                         1),
@@ -2105,17 +2114,17 @@ RETURN R1 1
 
     CHECK_EQ("\n" + compileFunction(R"(
 function test()
-	local i = 0
-	repeat
-		local j
+    local i = 0
+    repeat
+        local j
         j = i
-		foo(function() return j end)
-		i = i + 1
-		if bar then
-			break
-		end
-	until i < 5
-	return 0
+        foo(function() return j end)
+        i = i + 1
+        if bar then
+            break
+        end
+    until i < 5
+    return 0
 end
 )",
                         1),
@@ -2304,10 +2313,10 @@ local Value1, Value2, Value3 = ...
 local Table = {}
 
 Table.SubTable["Key"] = {
-	Key1 = Value1,
-	Key2 = Value2,
-	Key3 = Value3,
-	Key4 = true,
+    Key1 = Value1,
+    Key2 = Value2,
+    Key3 = Value3,
+    Key4 = true,
 }
 )");
 
@@ -2350,6 +2359,58 @@ Foo:Bar(
 )");
 }
 
+TEST_CASE("DebugLineInfoCallChain")
+{
+    Luau::BytecodeBuilder bcb;
+    bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code | Luau::BytecodeBuilder::Dump_Lines);
+    Luau::compileOrThrow(bcb, R"(
+local Foo = ...
+
+Foo
+:Bar(1)
+:Baz(2)
+.Qux(3)
+)");
+
+    CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
+2: GETVARARGS R0 1
+5: LOADN R4 1
+5: NAMECALL R2 R0 K0
+5: CALL R2 2 1
+6: LOADN R4 2
+6: NAMECALL R2 R2 K1
+6: CALL R2 2 1
+7: GETTABLEKS R1 R2 K2
+7: LOADN R2 3
+7: CALL R1 1 0
+8: RETURN R0 0
+)");
+}
+
+TEST_CASE("DebugLineInfoFastCall")
+{
+    Luau::BytecodeBuilder bcb;
+    bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code | Luau::BytecodeBuilder::Dump_Lines);
+    Luau::compileOrThrow(bcb, R"(
+local Foo, Bar = ...
+
+return
+    math.max(
+        Foo,
+        Bar)
+)");
+
+    CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
+2: GETVARARGS R0 2
+5: FASTCALL2 18 R0 R1 +5
+5: MOVE R3 R0
+5: MOVE R4 R1
+5: GETIMPORT R2 2
+5: CALL R2 2 -1
+5: RETURN R2 -1
+)");
+}
+
 TEST_CASE("DebugSource")
 {
     const char* source = R"(
@@ -2785,47 +2846,35 @@ CAPTURE UPVAL U1
 RETURN R0 1
 )");
 
-    if (FFlag::LuauPreloadClosuresUpval)
-    {
-        // recursive capture
-        CHECK_EQ("\n" + compileFunction("local function foo() return foo() end", 1), R"(
+    // recursive capture
+    CHECK_EQ("\n" + compileFunction("local function foo() return foo() end", 1), R"(
 DUPCLOSURE R0 K0
 CAPTURE VAL R0
 RETURN R0 0
 )");
 
-        // multi-level recursive capture
-        CHECK_EQ("\n" + compileFunction("local function foo() return function() return foo() end end", 1), R"(
+    // multi-level recursive capture
+    CHECK_EQ("\n" + compileFunction("local function foo() return function() return foo() end end", 1), R"(
 DUPCLOSURE R0 K0
 CAPTURE UPVAL U0
 RETURN R0 1
 )");
 
-        // multi-level recursive capture where function isn't top-level
-        // note: this should probably be optimized to DUPCLOSURE but doing that requires a different upval tracking flow in the compiler
-        CHECK_EQ("\n" + compileFunction(R"(
+    // multi-level recursive capture where function isn't top-level
+    // note: this should probably be optimized to DUPCLOSURE but doing that requires a different upval tracking flow in the compiler
+    CHECK_EQ("\n" + compileFunction(R"(
 local function foo()
     local function bar()
         return function() return bar() end
     end
 end
 )",
-                            1),
-            R"(
+                        1),
+        R"(
 NEWCLOSURE R0 P0
 CAPTURE UPVAL U0
 RETURN R0 1
 )");
-    }
-    else
-    {
-        // recursive capture
-        CHECK_EQ("\n" + compileFunction("local function foo() return foo() end", 1), R"(
-NEWCLOSURE R0 P0
-CAPTURE VAL R0
-RETURN R0 0
-)");
-    }
 }
 
 TEST_CASE("OutOfLocals")
@@ -3430,8 +3479,6 @@ local t = {
 
 TEST_CASE("ConstantClosure")
 {
-    ScopedFastFlag sff("LuauPreloadClosures", true);
-
     // closures without upvalues are created when bytecode is loaded
     CHECK_EQ("\n" + compileFunction(R"(
 return function() end
@@ -3467,15 +3514,13 @@ CAPTURE VAL R0
 RETURN R1 1
 )");
 
-    if (FFlag::LuauPreloadClosuresFenv)
-    {
-        // if they don't need upvalues but we sense that environment may be modified, we disable this to avoid fenv-related identity confusion
-        CHECK_EQ("\n" + compileFunction(R"(
+    // if they don't need upvalues but we sense that environment may be modified, we disable this to avoid fenv-related identity confusion
+    CHECK_EQ("\n" + compileFunction(R"(
 setfenv(1, {})
 return function() print("hi") end
 )",
-                            1),
-            R"(
+                        1),
+        R"(
 GETIMPORT R0 1
 LOADN R1 1
 NEWTABLE R2 0 0
@@ -3484,24 +3529,20 @@ NEWCLOSURE R0 P0
 RETURN R0 1
 )");
 
-        // note that fenv analysis isn't flow-sensitive right now, which is sort of a feature
-        CHECK_EQ("\n" + compileFunction(R"(
+    // note that fenv analysis isn't flow-sensitive right now, which is sort of a feature
+    CHECK_EQ("\n" + compileFunction(R"(
 if false then setfenv(1, {}) end
 return function() print("hi") end
 )",
-                            1),
-            R"(
+                        1),
+        R"(
 NEWCLOSURE R0 P0
 RETURN R0 1
 )");
-    }
 }
 
 TEST_CASE("SharedClosure")
 {
-    ScopedFastFlag sff1("LuauPreloadClosures", true);
-    ScopedFastFlag sff2("LuauPreloadClosuresUpval", true);
-
     // closures can be shared even if functions refer to upvalues, as long as upvalues are top-level
     CHECK_EQ("\n" + compileFunction(R"(
 local val = ...
@@ -3659,4 +3700,185 @@ RETURN R0 0
 )");
 }
 
+TEST_CASE("MutableGlobals")
+{
+    const char* source = R"(
+print()
+Game.print()
+Workspace.print()
+_G.print()
+game.print()
+plugin.print()
+script.print()
+shared.print()
+workspace.print()
+)";
+
+    // Check Roblox globals are no longer here
+    CHECK_EQ("\n" + compileFunction0(source), R"(
+GETIMPORT R0 1
+CALL R0 0 0
+GETIMPORT R0 3
+CALL R0 0 0
+GETIMPORT R0 5
+CALL R0 0 0
+GETIMPORT R1 7
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R0 9
+CALL R0 0 0
+GETIMPORT R0 11
+CALL R0 0 0
+GETIMPORT R0 13
+CALL R0 0 0
+GETIMPORT R0 15
+CALL R0 0 0
+GETIMPORT R0 17
+CALL R0 0 0
+RETURN R0 0
+)");
+
+    // Check we can add them back
+    Luau::BytecodeBuilder bcb;
+    bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code);
+    Luau::CompileOptions options;
+    const char* mutableGlobals[] = {"Game", "Workspace", "game", "plugin", "script", "shared", "workspace", NULL};
+    options.mutableGlobals = &mutableGlobals[0];
+    Luau::compileOrThrow(bcb, source, options);
+
+    CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
+GETIMPORT R0 1
+CALL R0 0 0
+GETIMPORT R1 3
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 5
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 7
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 9
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 11
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 13
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 15
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+GETIMPORT R1 17
+GETTABLEKS R0 R1 K0
+CALL R0 0 0
+RETURN R0 0
+)");
+}
+
+TEST_CASE("ConstantsNoFolding")
+{
+    const char* source = "return nil, true, 42, 'hello'";
+
+    Luau::BytecodeBuilder bcb;
+    bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code);
+    Luau::CompileOptions options;
+    options.optimizationLevel = 0;
+    Luau::compileOrThrow(bcb, source, options);
+
+    CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
+LOADNIL R0
+LOADB R1 1
+LOADK R2 K0
+LOADK R3 K1
+RETURN R0 4
+)");
+}
+
+TEST_CASE("VectorFastCall")
+{
+    const char* source = "return Vector3.new(1, 2, 3)";
+
+    Luau::BytecodeBuilder bcb;
+    bcb.setDumpFlags(Luau::BytecodeBuilder::Dump_Code);
+    Luau::CompileOptions options;
+    options.vectorLib = "Vector3";
+    options.vectorCtor = "new";
+    Luau::compileOrThrow(bcb, source, options);
+
+    CHECK_EQ("\n" + bcb.dumpFunction(0), R"(
+LOADN R1 1
+LOADN R2 2
+LOADN R3 3
+FASTCALL 54 +2
+GETIMPORT R0 2
+CALL R0 3 -1
+RETURN R0 -1
+)");
+}
+
+TEST_CASE("TypeAssertion")
+{
+    // validate that type assertions work with the compiler and that the code inside type assertion isn't evaluated
+    CHECK_EQ("\n" + compileFunction0(R"(
+print(foo() :: typeof(error("compile time")))
+)"),
+        R"(
+GETIMPORT R0 1
+GETIMPORT R1 3
+CALL R1 0 1
+CALL R0 1 0
+RETURN R0 0
+)");
+
+    // note that above, foo() is treated as single-arg function; removing type assertion changes the bytecode
+    CHECK_EQ("\n" + compileFunction0(R"(
+print(foo())
+)"),
+        R"(
+GETIMPORT R0 1
+GETIMPORT R1 3
+CALL R1 0 -1
+CALL R0 -1 0
+RETURN R0 0
+)");
+}
+
+TEST_CASE("Arithmetics")
+{
+    // basic arithmetics codegen with non-constants
+    CHECK_EQ("\n" + compileFunction0(R"(
+local a, b = ...
+return a + b, a - b, a / b, a * b, a % b, a ^ b
+)"),
+        R"(
+GETVARARGS R0 2
+ADD R2 R0 R1
+SUB R3 R0 R1
+DIV R4 R0 R1
+MUL R5 R0 R1
+MOD R6 R0 R1
+POW R7 R0 R1
+RETURN R2 6
+)");
+
+    // basic arithmetics codegen with constants on the right side
+    // note that we don't simplify these expressions as we don't know the type of a
+    CHECK_EQ("\n" + compileFunction0(R"(
+local a = ...
+return a + 1, a - 1, a / 1, a * 1, a % 1, a ^ 1
+)"),
+        R"(
+GETVARARGS R0 1
+ADDK R1 R0 K0
+SUBK R2 R0 K0
+DIVK R3 R0 K0
+MULK R4 R0 K0
+MODK R5 R0 K0
+POWK R6 R0 K0
+RETURN R1 6
+)");
+}
+
 TEST_SUITE_END();
