@@ -1,5 +1,7 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
-#include "Luau/Compiler.h"
+#include "lua.h"
+#include "lualib.h"
+#include "luacode.h"
 
 #include "Luau/BuiltinDefinitions.h"
 #include "Luau/ModuleResolver.h"
@@ -10,9 +12,6 @@
 #include "doctest.h"
 #include "ScopedFlags.h"
 
-#include "lua.h"
-#include "lualib.h"
-
 #include <fstream>
 #include <math.h>
 
@@ -49,8 +48,12 @@ static int lua_loadstring(lua_State* L)
 
     lua_setsafeenv(L, LUA_ENVIRONINDEX, false);
 
-    std::string bytecode = Luau::compile(std::string(s, l));
-    if (luau_load(L, chunkname, bytecode.data(), bytecode.size()) == 0)
+    size_t bytecodeSize = 0;
+    char* bytecode = luau_compile(s, l, nullptr, &bytecodeSize);
+    int result = luau_load(L, chunkname, bytecode, bytecodeSize, 0);
+    free(bytecode);
+
+    if (result == 0)
         return 1;
 
     lua_pushnil(L);
@@ -64,44 +67,42 @@ static int lua_vector(lua_State* L)
     double y = luaL_checknumber(L, 2);
     double z = luaL_checknumber(L, 3);
 
+#if LUA_VECTOR_SIZE == 4
+    double w = luaL_optnumber(L, 4, 0.0);
+    lua_pushvector(L, float(x), float(y), float(z), float(w));
+#else
     lua_pushvector(L, float(x), float(y), float(z));
+#endif
     return 1;
 }
 
 static int lua_vector_dot(lua_State* L)
 {
-    const float* a = lua_tovector(L, 1);
-    const float* b = lua_tovector(L, 2);
+    const float* a = luaL_checkvector(L, 1);
+    const float* b = luaL_checkvector(L, 2);
 
-    if (a && b)
-    {
-        lua_pushnumber(L, a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
-        return 1;
-    }
-
-    throw std::runtime_error("invalid arguments to vector:Dot");
+    lua_pushnumber(L, a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
+    return 1;
 }
 
 static int lua_vector_index(lua_State* L)
 {
+    const float* v = luaL_checkvector(L, 1);
     const char* name = luaL_checkstring(L, 2);
 
-    if (const float* v = lua_tovector(L, 1))
+    if (strcmp(name, "Magnitude") == 0)
     {
-        if (strcmp(name, "Magnitude") == 0)
-        {
-            lua_pushnumber(L, sqrtf(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
-            return 1;
-        }
-
-        if (strcmp(name, "Dot") == 0)
-        {
-            lua_pushcfunction(L, lua_vector_dot, "Dot");
-            return 1;
-        }
+        lua_pushnumber(L, sqrtf(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
+        return 1;
     }
 
-    throw std::runtime_error(Luau::format("%s is not a valid member of vector", name));
+    if (strcmp(name, "Dot") == 0)
+    {
+        lua_pushcfunction(L, lua_vector_dot, "Dot");
+        return 1;
+    }
+
+    luaL_error(L, "%s is not a valid member of vector", name);
 }
 
 static int lua_vector_namecall(lua_State* L)
@@ -112,7 +113,7 @@ static int lua_vector_namecall(lua_State* L)
             return lua_vector_dot(L);
     }
 
-    throw std::runtime_error(Luau::format("%s is not a valid method of vector", luaL_checkstring(L, 1)));
+    luaL_error(L, "%s is not a valid method of vector", luaL_checkstring(L, 1));
 }
 
 int lua_silence(lua_State* L)
@@ -122,8 +123,8 @@ int lua_silence(lua_State* L)
 
 using StateRef = std::unique_ptr<lua_State, void (*)(lua_State*)>;
 
-static StateRef runConformance(
-    const char* name, void (*setup)(lua_State* L) = nullptr, void (*yield)(lua_State* L) = nullptr, lua_State* initialLuaState = nullptr)
+static StateRef runConformance(const char* name, void (*setup)(lua_State* L) = nullptr, void (*yield)(lua_State* L) = nullptr,
+    lua_State* initialLuaState = nullptr, lua_CompileOptions* copts = nullptr)
 {
     std::string path = __FILE__;
     path.erase(path.find_last_of("\\/"));
@@ -179,21 +180,12 @@ static StateRef runConformance(
 
     std::string chunkname = "=" + std::string(name);
 
-    Luau::CompileOptions copts;
-    copts.debugLevel = 2;        // for debugger tests
-    copts.vectorCtor = "vector"; // for vector tests
+    size_t bytecodeSize = 0;
+    char* bytecode = luau_compile(source.data(), source.size(), copts, &bytecodeSize);
+    int result = luau_load(L, chunkname.c_str(), bytecode, bytecodeSize, 0);
+    free(bytecode);
 
-    std::string bytecode = Luau::compile(source, copts);
-    int status = 0;
-
-    if (luau_load(L, chunkname.c_str(), bytecode.data(), bytecode.size()) == 0)
-    {
-        status = lua_resume(L, nullptr, 0);
-    }
-    else
-    {
-        status = LUA_ERRSYNTAX;
-    }
+    int status = (result == 0) ? lua_resume(L, nullptr, 0) : LUA_ERRSYNTAX;
 
     while (yield && (status == LUA_YIELD || status == LUA_BREAK))
     {
@@ -240,8 +232,6 @@ TEST_CASE("Math")
 
 TEST_CASE("Table")
 {
-    ScopedFastFlag sff("LuauTableFreeze", true);
-
     runConformance("nextvar.lua");
 }
 
@@ -322,6 +312,8 @@ TEST_CASE("GC")
 
 TEST_CASE("Bitwise")
 {
+    ScopedFastFlag sff("LuauBit32Count", true);
+
     runConformance("bitwise.lua");
 }
 
@@ -332,57 +324,81 @@ TEST_CASE("UTF8")
 
 TEST_CASE("Coroutine")
 {
+    ScopedFastFlag sff("LuauCoroutineClose", true);
+
     runConformance("coroutine.lua");
 }
 
+static int cxxthrow(lua_State* L)
+{
+#if LUA_USE_LONGJMP
+    luaL_error(L, "oops");
+#else
+    throw std::runtime_error("oops");
+#endif
+}
+
 TEST_CASE("PCall")
 {
     runConformance("pcall.lua", [](lua_State* L) {
-        lua_pushcfunction(L, [](lua_State* L) -> int {
-#if LUA_USE_LONGJMP
-            luaL_error(L, "oops");
-#else
-            throw std::runtime_error("oops");
-#endif
-        });
+        lua_pushcfunction(L, cxxthrow, "cxxthrow");
         lua_setglobal(L, "cxxthrow");
 
-        lua_pushcfunction(L, [](lua_State* L) -> int {
-            lua_State* co = lua_tothread(L, 1);
-            lua_xmove(L, co, 1);
-            lua_resumeerror(co, L);
-            return 0;
-        });
+        lua_pushcfunction(
+            L,
+            [](lua_State* L) -> int {
+                lua_State* co = lua_tothread(L, 1);
+                lua_xmove(L, co, 1);
+                lua_resumeerror(co, L);
+                return 0;
+            },
+            "resumeerror");
         lua_setglobal(L, "resumeerror");
     });
 }
 
 TEST_CASE("Pack")
 {
+    ScopedFastFlag sff{"LuauStrPackUBCastFix", true};
+
     runConformance("tpack.lua");
 }
 
 TEST_CASE("Vector")
 {
-    runConformance("vector.lua", [](lua_State* L) {
-        lua_pushcfunction(L, lua_vector);
-        lua_setglobal(L, "vector");
+    ScopedFastFlag sff{"LuauIfElseExpressionBaseSupport", true};
 
-        lua_pushvector(L, 0.0f, 0.0f, 0.0f);
-        luaL_newmetatable(L, "vector");
+    lua_CompileOptions copts = {};
+    copts.optimizationLevel = 1;
+    copts.debugLevel = 1;
+    copts.vectorCtor = "vector";
 
-        lua_pushstring(L, "__index");
-        lua_pushcfunction(L, lua_vector_index);
-        lua_settable(L, -3);
+    runConformance(
+        "vector.lua",
+        [](lua_State* L) {
+            lua_pushcfunction(L, lua_vector, "vector");
+            lua_setglobal(L, "vector");
 
-        lua_pushstring(L, "__namecall");
-        lua_pushcfunction(L, lua_vector_namecall);
-        lua_settable(L, -3);
+#if LUA_VECTOR_SIZE == 4
+            lua_pushvector(L, 0.0f, 0.0f, 0.0f, 0.0f);
+#else
+            lua_pushvector(L, 0.0f, 0.0f, 0.0f);
+#endif
+            luaL_newmetatable(L, "vector");
 
-        lua_setreadonly(L, -1, true);
-        lua_setmetatable(L, -2);
-        lua_pop(L, 1);
-    });
+            lua_pushstring(L, "__index");
+            lua_pushcfunction(L, lua_vector_index, nullptr);
+            lua_settable(L, -3);
+
+            lua_pushstring(L, "__namecall");
+            lua_pushcfunction(L, lua_vector_namecall, nullptr);
+            lua_settable(L, -3);
+
+            lua_setreadonly(L, -1, true);
+            lua_setmetatable(L, -2);
+            lua_pop(L, 1);
+        },
+        nullptr, nullptr, &copts);
 }
 
 static void populateRTTI(lua_State* L, Luau::TypeId type)
@@ -486,6 +502,10 @@ TEST_CASE("Debugger")
     breakhits = 0;
     interruptedthread = nullptr;
 
+    lua_CompileOptions copts = {};
+    copts.optimizationLevel = 1;
+    copts.debugLevel = 2;
+
     runConformance(
         "debugger.lua",
         [](lua_State* L) {
@@ -495,6 +515,9 @@ TEST_CASE("Debugger")
             cb->debugbreak = [](lua_State* L, lua_Debug* ar) {
                 breakhits++;
 
+                // make sure we can trace the stack for every breakpoint we hit
+                lua_debugtrace(L);
+
                 // for every breakpoint, we break on the first invocation and continue on second
                 // this allows us to easily step off breakpoints
                 // (real implementaiton may require singlestepping)
@@ -511,15 +534,19 @@ TEST_CASE("Debugger")
             };
 
             // add breakpoint() function
-            lua_pushcfunction(L, [](lua_State* L) -> int {
-                int line = luaL_checkinteger(L, 1);
+            lua_pushcfunction(
+                L,
+                [](lua_State* L) -> int {
+                    int line = luaL_checkinteger(L, 1);
+                    bool enabled = luaL_optboolean(L, 2, true);
 
-                lua_Debug ar = {};
-                lua_getinfo(L, 1, "f", &ar);
+                    lua_Debug ar = {};
+                    lua_getinfo(L, 1, "f", &ar);
 
-                lua_breakpoint(L, -1, line, true);
-                return 0;
-            });
+                    lua_breakpoint(L, -1, line, enabled);
+                    return 0;
+                },
+                "breakpoint");
             lua_setglobal(L, "breakpoint");
         },
         [](lua_State* L) {
@@ -594,7 +621,8 @@ TEST_CASE("Debugger")
                 lua_resume(interruptedthread, nullptr, 0);
                 interruptedthread = nullptr;
             }
-        });
+        },
+        nullptr, &copts);
 
     CHECK(breakhits == 10); // 2 hits per breakpoint
 }
@@ -686,21 +714,52 @@ TEST_CASE("ApiFunctionCalls")
     StateRef globalState = runConformance("apicalls.lua");
     lua_State* L = globalState.get();
 
-    lua_getfield(L, LUA_GLOBALSINDEX, "add");
-    lua_pushnumber(L, 40);
-    lua_pushnumber(L, 2);
-    lua_call(L, 2, 1);
-    CHECK(lua_isnumber(L, -1));
-    CHECK(lua_tonumber(L, -1) == 42);
-    lua_pop(L, 1);
+    // lua_call
+    {
+        lua_getfield(L, LUA_GLOBALSINDEX, "add");
+        lua_pushnumber(L, 40);
+        lua_pushnumber(L, 2);
+        lua_call(L, 2, 1);
+        CHECK(lua_isnumber(L, -1));
+        CHECK(lua_tonumber(L, -1) == 42);
+        lua_pop(L, 1);
+    }
 
-    lua_getfield(L, LUA_GLOBALSINDEX, "add");
-    lua_pushnumber(L, 40);
-    lua_pushnumber(L, 2);
-    lua_pcall(L, 2, 1, 0);
-    CHECK(lua_isnumber(L, -1));
-    CHECK(lua_tonumber(L, -1) == 42);
-    lua_pop(L, 1);
+    // lua_pcall
+    {
+        lua_getfield(L, LUA_GLOBALSINDEX, "add");
+        lua_pushnumber(L, 40);
+        lua_pushnumber(L, 2);
+        lua_pcall(L, 2, 1, 0);
+        CHECK(lua_isnumber(L, -1));
+        CHECK(lua_tonumber(L, -1) == 42);
+        lua_pop(L, 1);
+    }
+
+    // lua_equal with a sleeping thread wake up
+    {
+        ScopedFastFlag luauActivateBeforeExec("LuauActivateBeforeExec", true);
+
+        lua_State* L2 = lua_newthread(L);
+
+        lua_getfield(L2, LUA_GLOBALSINDEX, "create_with_tm");
+        lua_pushnumber(L2, 42);
+        lua_pcall(L2, 1, 1, 0);
+
+        lua_getfield(L2, LUA_GLOBALSINDEX, "create_with_tm");
+        lua_pushnumber(L2, 42);
+        lua_pcall(L2, 1, 1, 0);
+
+        // Reset GC
+        lua_gc(L2, LUA_GCCOLLECT, 0);
+
+        // Try to mark 'L2' as sleeping
+        // Can't control GC precisely, even in tests
+        lua_gc(L2, LUA_GCSTEP, 8);
+
+        CHECK(lua_equal(L2, -1, -2) == 1);
+        lua_pop(L2, 2);
+    }
 }
 
 static bool endsWith(const std::string& str, const std::string& suffix)
@@ -714,8 +773,6 @@ static bool endsWith(const std::string& str, const std::string& suffix)
 #if !LUA_USE_LONGJMP
 TEST_CASE("ExceptionObject")
 {
-    ScopedFastFlag sff("LuauExceptionMessageFix", true);
-
     struct ExceptionResult
     {
         bool exceptionGenerated;
@@ -742,7 +799,7 @@ TEST_CASE("ExceptionObject")
         if (nsize == 0)
         {
             free(ptr);
-            return NULL;
+            return nullptr;
         }
         else if (nsize > 512 * 1024)
         {
@@ -801,4 +858,59 @@ TEST_CASE("IfElseExpression")
     runConformance("ifelseexpr.lua");
 }
 
+TEST_CASE("TagMethodError")
+{
+    ScopedFastFlag sff{"LuauCcallRestoreFix", true};
+
+    runConformance("tmerror.lua", [](lua_State* L) {
+        auto* cb = lua_callbacks(L);
+
+        cb->debugprotectederror = [](lua_State* L) {
+            CHECK(lua_isyieldable(L));
+        };
+    });
+}
+
+TEST_CASE("Coverage")
+{
+    lua_CompileOptions copts = {};
+    copts.optimizationLevel = 1;
+    copts.debugLevel = 1;
+    copts.coverageLevel = 2;
+
+    runConformance(
+        "coverage.lua",
+        [](lua_State* L) {
+            lua_pushcfunction(
+                L,
+                [](lua_State* L) -> int {
+                    luaL_argexpected(L, lua_isLfunction(L, 1), 1, "function");
+
+                    lua_newtable(L);
+                    lua_getcoverage(L, 1, L, [](void* context, const char* function, int linedefined, int depth, const int* hits, size_t size) {
+                        lua_State* L = static_cast<lua_State*>(context);
+
+                        lua_newtable(L);
+
+                        lua_pushstring(L, function);
+                        lua_setfield(L, -2, "name");
+
+                        for (size_t i = 0; i < size; ++i)
+                            if (hits[i] != -1)
+                            {
+                                lua_pushinteger(L, hits[i]);
+                                lua_rawseti(L, -2, int(i));
+                            }
+
+                        lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
+                    });
+
+                    return 1;
+                },
+                "getcoverage");
+            lua_setglobal(L, "getcoverage");
+        },
+        nullptr, nullptr, &copts);
+}
+
 TEST_SUITE_END();
