@@ -7,6 +7,8 @@
 
 #include "doctest.h"
 
+LUAU_FASTFLAG(LuauFixTonumberReturnType)
+
 using namespace Luau;
 
 TEST_SUITE_BEGIN("BuiltinTests");
@@ -359,7 +361,7 @@ TEST_CASE_FIXTURE(Fixture, "table_insert_correctly_infers_type_of_array_2_args_o
     CHECK_EQ(typeChecker.stringType, requireType("s"));
 }
 
-TEST_CASE_FIXTURE(Fixture, "table_insert_corrrectly_infers_type_of_array_3_args_overload")
+TEST_CASE_FIXTURE(Fixture, "table_insert_correctly_infers_type_of_array_3_args_overload")
 {
     CheckResult result = check(R"(
         local t = {}
@@ -444,8 +446,6 @@ TEST_CASE_FIXTURE(Fixture, "os_time_takes_optional_date_table")
 
 TEST_CASE_FIXTURE(Fixture, "thread_is_a_type")
 {
-    ScopedFastFlag sff{"LuauDontMutatePersistentFunctions", true};
-
     CheckResult result = check(R"(
         local co = coroutine.create(function() end)
     )");
@@ -456,8 +456,6 @@ TEST_CASE_FIXTURE(Fixture, "thread_is_a_type")
 
 TEST_CASE_FIXTURE(Fixture, "coroutine_resume_anything_goes")
 {
-    ScopedFastFlag sff{"LuauDontMutatePersistentFunctions", true};
-
     CheckResult result = check(R"(
         local function nifty(x, y)
             print(x, y)
@@ -476,8 +474,6 @@ TEST_CASE_FIXTURE(Fixture, "coroutine_resume_anything_goes")
 
 TEST_CASE_FIXTURE(Fixture, "coroutine_wrap_anything_goes")
 {
-    ScopedFastFlag sff{"LuauDontMutatePersistentFunctions", true};
-
     CheckResult result = check(R"(
         --!nonstrict
         local function nifty(x, y)
@@ -820,10 +816,32 @@ TEST_CASE_FIXTURE(Fixture, "string_format_report_all_type_errors_at_correct_posi
     CHECK_EQ(TypeErrorData(TypeMismatch{stringType, booleanType}), result.errors[2].data);
 }
 
+TEST_CASE_FIXTURE(Fixture, "tonumber_returns_optional_number_type")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local b: number = tonumber('asdf')
+    )");
+
+    if (FFlag::LuauFixTonumberReturnType)
+    {
+        LUAU_REQUIRE_ERROR_COUNT(1, result);
+        CHECK_EQ("Type 'number?' could not be converted into 'number'", toString(result.errors[0]));
+    }
+}
+
+TEST_CASE_FIXTURE(Fixture, "tonumber_returns_optional_number_type2")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local b: number = tonumber('asdf') or 1
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
 TEST_CASE_FIXTURE(Fixture, "dont_add_definitions_to_persistent_types")
 {
-    ScopedFastFlag sff{"LuauDontMutatePersistentFunctions", true};
-
     CheckResult result = check(R"(
         local f = math.sin
         local function g(x) return math.sin(x) end
