@@ -9,12 +9,13 @@
 #include <algorithm>
 
 LUAU_FASTFLAG(LuauEqConstraint)
+LUAU_FASTFLAG(LuauQuantifyInPlace2)
 
 using namespace Luau;
 
 TEST_SUITE_BEGIN("ProvisionalTests");
 
-// These tests check for behavior that differes from the final behavior we'd
+// These tests check for behavior that differs from the final behavior we'd
 // like to have.  They serve to document the current state of the typechecker.
 // When making future improvements, its very likely these tests will break and
 // will need to be replaced.
@@ -40,16 +41,30 @@ TEST_CASE_FIXTURE(Fixture, "typeguard_inference_incomplete")
         end
     )";
 
-    const std::string expected = R"(
-        function f(a:{fn:()->(free)}): ()
+    const std::string old_expected = R"(
+        function f(a:{fn:()->(free,free...)}): ()
             if type(a) == 'boolean'then
                 local a1:boolean=a
             elseif a.fn()then
-                local a2:{fn:()->(free)}=a
+                local a2:{fn:()->(free,free...)}=a
             end
         end
     )";
-    CHECK_EQ(expected, decorateWithTypes(code));
+
+    const std::string expected = R"(
+        function f(a:{fn:()->(a,b...)}): ()
+            if type(a) == 'boolean'then
+                local a1:boolean=a
+            elseif a.fn()then
+                local a2:{fn:()->(a,b...)}=a
+            end
+        end
+    )";
+
+    if (FFlag::LuauQuantifyInPlace2)
+        CHECK_EQ(expected, decorateWithTypes(code));
+    else
+        CHECK_EQ(old_expected, decorateWithTypes(code));
 }
 
 TEST_CASE_FIXTURE(Fixture, "xpcall_returns_what_f_returns")
@@ -177,9 +192,6 @@ TEST_CASE_FIXTURE(Fixture, "normal_conditional_expression_has_refinements")
 // Luau currently doesn't yet know how to allow assignments when the binding was refined.
 TEST_CASE_FIXTURE(Fixture, "while_body_are_also_refined")
 {
-    ScopedFastFlag sffs2{"LuauGenericFunctions", true};
-    ScopedFastFlag sffs5{"LuauParseGenericFunctions", true};
-
     CheckResult result = check(R"(
         type Node<T> = { value: T, child: Node<T>? }
 
@@ -231,16 +243,7 @@ TEST_CASE_FIXTURE(Fixture, "operator_eq_completely_incompatible")
         local r2 = b == a
     )");
 
-    if (FFlag::LuauEqConstraint)
-    {
-        LUAU_REQUIRE_NO_ERRORS(result);
-    }
-    else
-    {
-        LUAU_REQUIRE_ERROR_COUNT(2, result);
-        CHECK_EQ(toString(result.errors[0]), "Type '{| x: string |}?' could not be converted into 'number | string'");
-        CHECK_EQ(toString(result.errors[1]), "Type 'number | string' could not be converted into '{| x: string |}?'");
-    }
+    LUAU_REQUIRE_NO_ERRORS(result);
 }
 
 // Belongs in TypeInfer.refinements.test.cpp.
@@ -270,8 +273,8 @@ TEST_CASE_FIXTURE(Fixture, "lvalue_equals_another_lvalue_with_no_overlap")
 
 TEST_CASE_FIXTURE(Fixture, "bail_early_if_unification_is_too_complicated" * doctest::timeout(0.5))
 {
-    ScopedFastInt sffi{"LuauTarjanChildLimit", 50};
-    ScopedFastInt sffi2{"LuauTypeInferIterationLimit", 50};
+    ScopedFastInt sffi{"LuauTarjanChildLimit", 1};
+    ScopedFastInt sffi2{"LuauTypeInferIterationLimit", 1};
 
     CheckResult result = check(R"LUA(
         local Result
@@ -542,6 +545,25 @@ TEST_CASE_FIXTURE(Fixture, "bail_early_on_typescript_port_of_Result_type" * doct
     }
 }
 
+TEST_CASE_FIXTURE(Fixture, "table_subtyping_shouldn't_add_optional_properties_to_sealed_tables")
+{
+    CheckResult result = check(R"(
+        --!strict
+        local function setNumber(t: { p: number? }, x:number) t.p = x end
+        local function getString(t: { p: string? }):string return t.p or "" end
+        -- This shouldn't type-check!
+        local function oh(x:number): string
+          local t: {} = {}
+          setNumber(t, x)
+          return getString(t)
+        end
+        local s: string = oh(37)
+    )");
+
+    // Really this should return an error, but it doesn't
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
 // Should be in TypeInfer.tables.test.cpp
 // It's unsound to instantiate tables containing generic methods,
 // since mutating properties means table properties should be invariant.
@@ -569,11 +591,9 @@ TEST_CASE_FIXTURE(Fixture, "invariant_table_properties_means_instantiating_table
 TEST_CASE_FIXTURE(Fixture, "self_recursive_instantiated_param")
 {
     ScopedFastFlag luauCloneCorrectlyBeforeMutatingTableType{"LuauCloneCorrectlyBeforeMutatingTableType", true};
-    ScopedFastFlag luauFollowInTypeFunApply{"LuauFollowInTypeFunApply", true};
-    ScopedFastFlag luauInstantiatedTypeParamRecursion{"LuauInstantiatedTypeParamRecursion", true};
 
     // Mutability in type function application right now can create strange recursive types
-    // TODO: instantiation right now is problematic, it this example should either leave the Table type alone
+    // TODO: instantiation right now is problematic, in this example should either leave the Table type alone
     // or it should rename the type to 'Self' so that the result will be 'Self<Table>'
     CheckResult result = check(R"(
 type Table = { a: number }
