@@ -1,4 +1,5 @@
 // This file is part of the Luau programming language and is licensed under MIT License; see LICENSE.txt for details
+#include "Luau/Scope.h"
 #include "Luau/TypeInfer.h"
 
 #include "Fixture.h"
@@ -6,8 +7,7 @@
 #include "doctest.h"
 
 LUAU_FASTFLAG(LuauWeakEqConstraint)
-LUAU_FASTFLAG(LuauImprovedTypeGuardPredicate2)
-LUAU_FASTFLAG(LuauOrPredicate)
+LUAU_FASTFLAG(LuauQuantifyInPlace2)
 
 using namespace Luau;
 
@@ -132,11 +132,8 @@ TEST_CASE_FIXTURE(Fixture, "or_predicate_with_truthy_predicates")
     CHECK_EQ("string?", toString(requireTypeAtPosition({3, 26})));
     CHECK_EQ("number?", toString(requireTypeAtPosition({4, 26})));
 
-    if (FFlag::LuauOrPredicate)
-    {
-        CHECK_EQ("nil", toString(requireTypeAtPosition({6, 26})));
-        CHECK_EQ("nil", toString(requireTypeAtPosition({7, 26})));
-    }
+    CHECK_EQ("nil", toString(requireTypeAtPosition({6, 26})));
+    CHECK_EQ("nil", toString(requireTypeAtPosition({7, 26})));
 }
 
 TEST_CASE_FIXTURE(Fixture, "type_assertion_expr_carry_its_constraints")
@@ -199,16 +196,8 @@ TEST_CASE_FIXTURE(Fixture, "typeguard_only_look_up_types_from_global_scope")
         end
     )");
 
-    if (FFlag::LuauImprovedTypeGuardPredicate2)
-    {
-        LUAU_REQUIRE_ERROR_COUNT(1, result);
-        CHECK_EQ("Type 'number' has no overlap with 'string'", toString(result.errors[0]));
-    }
-    else
-    {
-        LUAU_REQUIRE_ERROR_COUNT(1, result);
-        CHECK_EQ("Type 'string' could not be converted into 'boolean'", toString(result.errors[0]));
-    }
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ("Type 'number' has no overlap with 'string'", toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "call_a_more_specific_function_using_typeguard")
@@ -290,6 +279,8 @@ TEST_CASE_FIXTURE(Fixture, "assert_non_binary_expressions_actually_resolve_const
 
 TEST_CASE_FIXTURE(Fixture, "assign_table_with_refined_property_with_a_similar_type_is_illegal")
 {
+    ScopedFastFlag LuauTableSubtypingVariance2{"LuauTableSubtypingVariance2", true};
+    ScopedFastFlag luauExtendedTypeMismatchError{"LuauExtendedTypeMismatchError", true};
 
     CheckResult result = check(R"(
         local t: {x: number?} = {x = nil}
@@ -300,7 +291,10 @@ TEST_CASE_FIXTURE(Fixture, "assign_table_with_refined_property_with_a_similar_ty
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
-    CHECK_EQ("Type '{| x: number? |}' could not be converted into '{| x: number |}'", toString(result.errors[0]));
+    CHECK_EQ(R"(Type '{| x: number? |}' could not be converted into '{| x: number |}'
+caused by:
+  Property 'x' is not compatible. Type 'number?' could not be converted into 'number')",
+        toString(result.errors[0]));
 }
 
 TEST_CASE_FIXTURE(Fixture, "lvalue_is_equal_to_another_lvalue")
@@ -526,8 +520,6 @@ TEST_CASE_FIXTURE(Fixture, "narrow_property_of_a_bounded_variable")
 
 TEST_CASE_FIXTURE(Fixture, "type_narrow_to_vector")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(x)
             if type(x) == "vector" then
@@ -544,8 +536,6 @@ TEST_CASE_FIXTURE(Fixture, "type_narrow_to_vector")
 
 TEST_CASE_FIXTURE(Fixture, "nonoptional_type_can_narrow_to_nil_if_sense_is_true")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local t = {"hello"}
         local v = t[2]
@@ -573,8 +563,6 @@ TEST_CASE_FIXTURE(Fixture, "nonoptional_type_can_narrow_to_nil_if_sense_is_true"
 
 TEST_CASE_FIXTURE(Fixture, "typeguard_not_to_be_string")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(x: string | number | boolean)
             if type(x) ~= "string" then
@@ -593,8 +581,6 @@ TEST_CASE_FIXTURE(Fixture, "typeguard_not_to_be_string")
 
 TEST_CASE_FIXTURE(Fixture, "typeguard_narrows_for_table")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(x: string | {x: number} | {y: boolean})
             if type(x) == "table" then
@@ -613,8 +599,6 @@ TEST_CASE_FIXTURE(Fixture, "typeguard_narrows_for_table")
 
 TEST_CASE_FIXTURE(Fixture, "typeguard_narrows_for_functions")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function weird(x: string | ((number) -> string))
             if type(x) == "function" then
@@ -698,8 +682,6 @@ struct RefinementClassFixture : Fixture
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "typeguard_cast_free_table_to_vector")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(vec)
             local X, Y, Z = vec.X, vec.Y, vec.Z
@@ -718,16 +700,21 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "typeguard_cast_free_table_to_vector")
 
     CHECK_EQ("Vector3", toString(requireTypeAtPosition({5, 28}))); // type(vec) == "vector"
 
-    CHECK_EQ("Type '{- X: a, Y: b, Z: c -}' could not be converted into 'Instance'", toString(result.errors[0]));
+    if (FFlag::LuauQuantifyInPlace2)
+        CHECK_EQ("Type '{+ X: a, Y: b, Z: c +}' could not be converted into 'Instance'", toString(result.errors[0]));
+    else
+        CHECK_EQ("Type '{- X: a, Y: b, Z: c -}' could not be converted into 'Instance'", toString(result.errors[0]));
+
     CHECK_EQ("*unknown*", toString(requireTypeAtPosition({7, 28}))); // typeof(vec) == "Instance"
 
-    CHECK_EQ("{- X: a, Y: b, Z: c -}", toString(requireTypeAtPosition({9, 28}))); // type(vec) ~= "vector" and typeof(vec) ~= "Instance"
+    if (FFlag::LuauQuantifyInPlace2)
+        CHECK_EQ("{+ X: a, Y: b, Z: c +}", toString(requireTypeAtPosition({9, 28}))); // type(vec) ~= "vector" and typeof(vec) ~= "Instance"
+    else
+        CHECK_EQ("{- X: a, Y: b, Z: c -}", toString(requireTypeAtPosition({9, 28}))); // type(vec) ~= "vector" and typeof(vec) ~= "Instance"
 }
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "typeguard_cast_instance_or_vector3_to_vector")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(x: Instance | Vector3)
             if typeof(x) == "Vector3" then
@@ -746,8 +733,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "typeguard_cast_instance_or_vector3_to
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "type_narrow_for_all_the_userdata")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(x: string | number | Instance | Vector3)
             if type(x) == "userdata" then
@@ -766,11 +751,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "type_narrow_for_all_the_userdata")
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "eliminate_subclasses_of_instance")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauImprovedTypeGuardPredicate2", true},
-        {"LuauTypeGuardPeelsAwaySubclasses", true},
-    };
-
     CheckResult result = check(R"(
         local function f(x: Part | Folder | string)
             if typeof(x) == "Instance" then
@@ -789,11 +769,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "eliminate_subclasses_of_instance")
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "narrow_this_large_union")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauImprovedTypeGuardPredicate2", true},
-        {"LuauTypeGuardPeelsAwaySubclasses", true},
-    };
-
     CheckResult result = check(R"(
         local function f(x: Part | Folder | Instance | string | Vector3 | any)
             if typeof(x) == "Instance" then
@@ -812,11 +787,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "narrow_this_large_union")
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "x_as_any_if_x_is_instance_elseif_x_is_table")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-    };
-
     CheckResult result = check(R"(
         --!nonstrict
 
@@ -837,12 +807,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "x_as_any_if_x_is_instance_elseif_x_is
 
 TEST_CASE_FIXTURE(RefinementClassFixture, "x_is_not_instance_or_else_not_part")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-        {"LuauTypeGuardPeelsAwaySubclasses", true},
-    };
-
     CheckResult result = check(R"(
         local function f(x: Part | Folder | string)
             if typeof(x) ~= "Instance" or not x:IsA("Part") then
@@ -861,8 +825,6 @@ TEST_CASE_FIXTURE(RefinementClassFixture, "x_is_not_instance_or_else_not_part")
 
 TEST_CASE_FIXTURE(Fixture, "type_guard_can_filter_for_intersection_of_tables")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         type XYCoord = {x: number} & {y: number}
         local function f(t: XYCoord?)
@@ -882,8 +844,6 @@ TEST_CASE_FIXTURE(Fixture, "type_guard_can_filter_for_intersection_of_tables")
 
 TEST_CASE_FIXTURE(Fixture, "type_guard_can_filter_for_overloaded_function")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         type SomeOverloadedFunction = ((number) -> string) & ((string) -> number)
         local function f(g: SomeOverloadedFunction?)
@@ -903,8 +863,6 @@ TEST_CASE_FIXTURE(Fixture, "type_guard_can_filter_for_overloaded_function")
 
 TEST_CASE_FIXTURE(Fixture, "type_guard_warns_on_no_overlapping_types_only_when_sense_is_true")
 {
-    ScopedFastFlag sff2{"LuauImprovedTypeGuardPredicate2", true};
-
     CheckResult result = check(R"(
         local function f(t: {x: number})
             if type(t) ~= "table" then
@@ -923,8 +881,6 @@ TEST_CASE_FIXTURE(Fixture, "type_guard_warns_on_no_overlapping_types_only_when_s
 
 TEST_CASE_FIXTURE(Fixture, "not_a_or_not_b")
 {
-    ScopedFastFlag sff{"LuauOrPredicate", true};
-
     CheckResult result = check(R"(
         local function f(a: number?, b: number?)
             if (not a) or (not b) then
@@ -942,8 +898,6 @@ TEST_CASE_FIXTURE(Fixture, "not_a_or_not_b")
 
 TEST_CASE_FIXTURE(Fixture, "not_a_or_not_b2")
 {
-    ScopedFastFlag sff{"LuauOrPredicate", true};
-
     CheckResult result = check(R"(
         local function f(a: number?, b: number?)
             if not (a and b) then
@@ -961,8 +915,6 @@ TEST_CASE_FIXTURE(Fixture, "not_a_or_not_b2")
 
 TEST_CASE_FIXTURE(Fixture, "not_a_and_not_b")
 {
-    ScopedFastFlag sff{"LuauOrPredicate", true};
-
     CheckResult result = check(R"(
         local function f(a: number?, b: number?)
             if (not a) and (not b) then
@@ -980,8 +932,6 @@ TEST_CASE_FIXTURE(Fixture, "not_a_and_not_b")
 
 TEST_CASE_FIXTURE(Fixture, "not_a_and_not_b2")
 {
-    ScopedFastFlag sff{"LuauOrPredicate", true};
-
     CheckResult result = check(R"(
         local function f(a: number?, b: number?)
             if not (a or b) then
@@ -999,11 +949,6 @@ TEST_CASE_FIXTURE(Fixture, "not_a_and_not_b2")
 
 TEST_CASE_FIXTURE(Fixture, "either_number_or_string")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-    };
-
     CheckResult result = check(R"(
         local function f(x: any)
             if type(x) == "number" or type(x) == "string" then
@@ -1019,8 +964,6 @@ TEST_CASE_FIXTURE(Fixture, "either_number_or_string")
 
 TEST_CASE_FIXTURE(Fixture, "not_t_or_some_prop_of_t")
 {
-    ScopedFastFlag sff{"LuauOrPredicate", true};
-
     CheckResult result = check(R"(
         local function f(t: {x: boolean}?)
             if not t or t.x then
@@ -1036,11 +979,6 @@ TEST_CASE_FIXTURE(Fixture, "not_t_or_some_prop_of_t")
 
 TEST_CASE_FIXTURE(Fixture, "assert_a_to_be_truthy_then_assert_a_to_be_number")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-    };
-
     CheckResult result = check(R"(
         local a: (number | string)?
         assert(a)
@@ -1057,11 +995,6 @@ TEST_CASE_FIXTURE(Fixture, "assert_a_to_be_truthy_then_assert_a_to_be_number")
 
 TEST_CASE_FIXTURE(Fixture, "merge_should_be_fully_agnostic_of_hashmap_ordering")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-    };
-
     // This bug came up because there was a mistake in Luau::merge where zipping on two maps would produce the wrong merged result.
     CheckResult result = check(R"(
         local function f(b: string | { x: string }, a)
@@ -1081,11 +1014,6 @@ TEST_CASE_FIXTURE(Fixture, "merge_should_be_fully_agnostic_of_hashmap_ordering")
 
 TEST_CASE_FIXTURE(Fixture, "refine_the_correct_types_opposite_of_when_a_is_not_number_or_string")
 {
-    ScopedFastFlag sffs[] = {
-        {"LuauOrPredicate", true},
-        {"LuauImprovedTypeGuardPredicate2", true},
-    };
-
     CheckResult result = check(R"(
         local function f(a: string | number | boolean)
             if type(a) ~= "number" and type(a) ~= "string" then
@@ -1157,4 +1085,19 @@ TEST_CASE_FIXTURE(Fixture, "type_comparison_ifelse_expression")
     CHECK_EQ("any", toString(requireTypeAtPosition({6, 66})));
 }
 
+TEST_CASE_FIXTURE(Fixture, "apply_refinements_on_astexprindexexpr_whose_subscript_expr_is_constant_string")
+{
+    ScopedFastFlag sff{"LuauRefiLookupFromIndexExpr", true};
+
+    CheckResult result = check(R"(
+        type T = { [string]: { prop: number }? }
+        local t: T = {}
+        if t["hello"] then
+            local foo = t["hello"].prop
+        end
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
 TEST_SUITE_END();
