@@ -12,6 +12,8 @@
 
 using namespace Luau;
 
+LUAU_FASTFLAG(LuauExtendedFunctionMismatchError)
+
 TEST_SUITE_BEGIN("TableTests");
 
 TEST_CASE_FIXTURE(Fixture, "basic")
@@ -46,6 +48,21 @@ TEST_CASE_FIXTURE(Fixture, "augment_table")
     CHECK(tType->props.find("foo") != tType->props.end());
 }
 
+TEST_CASE_FIXTURE(Fixture, "augment_nested_table")
+{
+    CheckResult result = check("local t = { p = {} }  t.p.foo = 'bar'");
+    LUAU_REQUIRE_NO_ERRORS(result);
+
+    TableTypeVar* tType = getMutable<TableTypeVar>(requireType("t"));
+    REQUIRE(tType != nullptr);
+
+    REQUIRE(tType->props.find("p") != tType->props.end());
+    const TableTypeVar* pType = get<TableTypeVar>(tType->props["p"].type);
+    REQUIRE(pType != nullptr);
+
+    CHECK(pType->props.find("foo") != pType->props.end());
+}
+
 TEST_CASE_FIXTURE(Fixture, "cannot_augment_sealed_table")
 {
     CheckResult result = check("local t = {prop=999}    t.foo = 'bar'");
@@ -260,6 +277,8 @@ TEST_CASE_FIXTURE(Fixture, "open_table_unification")
 
 TEST_CASE_FIXTURE(Fixture, "open_table_unification_2")
 {
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
     CheckResult result = check(R"(
         local a = {}
         a.x = 99
@@ -272,10 +291,11 @@ TEST_CASE_FIXTURE(Fixture, "open_table_unification_2")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     TypeError& err = result.errors[0];
-    UnknownProperty* error = get<UnknownProperty>(err);
+    MissingProperties* error = get<MissingProperties>(err);
     REQUIRE(error != nullptr);
+    REQUIRE(error->properties.size() == 1);
 
-    CHECK_EQ(error->key, "y");
+    CHECK_EQ("y", error->properties[0]);
     // TODO(rblanckaert): Revist when we can bind self at function creation time
     // CHECK_EQ(err.location, Location(Position{5, 19}, Position{5, 25}));
 
@@ -328,6 +348,8 @@ TEST_CASE_FIXTURE(Fixture, "table_param_row_polymorphism_1")
 
 TEST_CASE_FIXTURE(Fixture, "table_param_row_polymorphism_2")
 {
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
     CheckResult result = check(R"(
         --!strict
         function foo(o)
@@ -340,14 +362,17 @@ TEST_CASE_FIXTURE(Fixture, "table_param_row_polymorphism_2")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    UnknownProperty* error = get<UnknownProperty>(result.errors[0]);
+    MissingProperties* error = get<MissingProperties>(result.errors[0]);
     REQUIRE(error != nullptr);
+    REQUIRE(error->properties.size() == 1);
 
-    CHECK_EQ("baz", error->key);
+    CHECK_EQ("baz", error->properties[0]);
 }
 
 TEST_CASE_FIXTURE(Fixture, "table_param_row_polymorphism_3")
 {
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
     CheckResult result = check(R"(
         local T = {}
         T.bar = 'hello'
@@ -359,8 +384,11 @@ TEST_CASE_FIXTURE(Fixture, "table_param_row_polymorphism_3")
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
     TypeError& err = result.errors[0];
-    UnknownProperty* error = get<UnknownProperty>(err);
+    MissingProperties* error = get<MissingProperties>(err);
     REQUIRE(error != nullptr);
+    REQUIRE(error->properties.size() == 1);
+
+    CHECK_EQ("baz", error->properties[0]);
 
     // TODO(rblanckaert): Revist when we can bind self at function creation time
     /*
@@ -448,6 +476,73 @@ TEST_CASE_FIXTURE(Fixture, "ok_to_add_property_to_free_table")
     dumpErrors(result);
 }
 
+TEST_CASE_FIXTURE(Fixture, "okay_to_add_property_to_unsealed_tables_by_assignment")
+{
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local t = { u = {} }
+        t = { u = { p = 37 } }
+        t = { u = { q = "hi" } }
+        local x = t.u.p
+        local y = t.u.q
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+    CHECK_EQ("number?", toString(requireType("x")));
+    CHECK_EQ("string?", toString(requireType("y")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "okay_to_add_property_to_unsealed_tables_by_function_call")
+{
+    CheckResult result = check(R"(
+        --!strict
+        function get(x) return x.opts["MYOPT"] end
+        function set(x,y) x.opts["MYOPT"] = y end
+        local t = { opts = {} }
+        set(t,37)
+        local x = get(t)
+    )");
+
+    // Currently this errors but it shouldn't, since set only needs write access
+    // TODO: file a JIRA for this
+    LUAU_REQUIRE_ERRORS(result);
+    // CHECK_EQ("number?", toString(requireType("x")));
+}
+
+TEST_CASE_FIXTURE(Fixture, "width_subtyping")
+{
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
+    CheckResult result = check(R"(
+        --!strict
+        function f(x : { q : number })
+           x.q = 8
+        end
+        local t : { q : number, r : string } = { q = 8, r = "hi" }
+        f(t)
+        local x : string = t.r
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "width_subtyping_needs_covariance")
+{
+    CheckResult result = check(R"(
+        --!strict
+        function f(x : { p : { q : number }})
+           x.p = { q = 8, r = 5 }
+        end
+        local t : { p : { q : number, r : string } } = { p = { q = 8, r = "hi" } }
+        f(t) -- Shouldn't typecheck
+        local x : string = t.p.r -- x is 5
+    )");
+
+    LUAU_REQUIRE_ERRORS(result);
+}
+
 TEST_CASE_FIXTURE(Fixture, "infer_array")
 {
     CheckResult result = check(R"(
@@ -524,7 +619,7 @@ TEST_CASE_FIXTURE(Fixture, "indexers_get_quantified_too")
 
     REQUIRE_EQ(indexer.indexType, typeChecker.numberType);
 
-    REQUIRE(nullptr != get<GenericTypeVar>(indexer.indexResultType));
+    REQUIRE(nullptr != get<GenericTypeVar>(follow(indexer.indexResultType)));
 }
 
 TEST_CASE_FIXTURE(Fixture, "indexers_quantification_2")
@@ -676,16 +771,27 @@ TEST_CASE_FIXTURE(Fixture, "infer_indexer_for_left_unsealed_table_from_right_han
     LUAU_REQUIRE_NO_ERRORS(result);
 }
 
-TEST_CASE_FIXTURE(Fixture, "sealed_table_value_must_not_infer_an_indexer")
+TEST_CASE_FIXTURE(Fixture, "sealed_table_value_can_infer_an_indexer")
 {
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
     CheckResult result = check(R"(
         local t: { a: string, [number]: string } = { a = "foo" }
     )");
 
-    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
 
-    TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
-    REQUIRE(tm != nullptr);
+TEST_CASE_FIXTURE(Fixture, "array_factory_function")
+{
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
+    CheckResult result = check(R"(
+        function empty() return {} end
+        local array: {string} = empty()
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
 }
 
 TEST_CASE_FIXTURE(Fixture, "sealed_table_indexers_must_unify")
@@ -756,37 +862,6 @@ TEST_CASE_FIXTURE(Fixture, "indexing_from_a_table_should_prefer_properties_when_
     CHECK_MESSAGE(nullptr != get<TypeMismatch>(result.errors[0]), "Expected a TypeMismatch but got " << result.errors[0]);
 }
 
-TEST_CASE_FIXTURE(Fixture, "indexing_from_a_table_with_a_string")
-{
-    ScopedFastFlag fflag("LuauIndexTablesWithIndexers", true);
-
-    CheckResult result = check(R"(
-        local t: { a: string }
-        function f(x: string) return t[x] end
-        local a = f("a")
-        local b = f("b")
-    )");
-
-    LUAU_REQUIRE_NO_ERRORS(result);
-
-    CHECK_EQ(*typeChecker.anyType, *requireType("a"));
-    CHECK_EQ(*typeChecker.anyType, *requireType("b"));
-}
-
-TEST_CASE_FIXTURE(Fixture, "indexing_from_a_table_with_a_number")
-{
-    ScopedFastFlag fflag("LuauIndexTablesWithIndexers", true);
-
-    CheckResult result = check(R"(
-        local t = { a = true }
-        function f(x: number) return t[x] end
-    )");
-
-    LUAU_REQUIRE_ERROR_COUNT(1, result);
-
-    CHECK_MESSAGE(nullptr != get<TypeMismatch>(result.errors[0]), "Expected a TypeMismatch but got " << result.errors[0]);
-}
-
 TEST_CASE_FIXTURE(Fixture, "assigning_to_an_unsealed_table_with_string_literal_should_infer_new_properties_over_indexer")
 {
     CheckResult result = check(R"(
@@ -1392,6 +1467,8 @@ TEST_CASE_FIXTURE(Fixture, "casting_tables_with_props_into_table_with_indexer2")
 
 TEST_CASE_FIXTURE(Fixture, "casting_tables_with_props_into_table_with_indexer3")
 {
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
     CheckResult result = check(R"(
         local function foo(a: {[string]: number, a: string}) end
         foo({ a = 1 })
@@ -1402,8 +1479,21 @@ TEST_CASE_FIXTURE(Fixture, "casting_tables_with_props_into_table_with_indexer3")
     ToStringOptions o{/* exhaustive= */ true};
     TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
     REQUIRE(tm);
-    CHECK_EQ("string", toString(tm->wantedType, o));
-    CHECK_EQ("number", toString(tm->givenType, o));
+    CHECK_EQ("{| [string]: number, a: string |}", toString(tm->wantedType, o));
+    CHECK_EQ("{| a: number |}", toString(tm->givenType, o));
+}
+
+TEST_CASE_FIXTURE(Fixture, "casting_tables_with_props_into_table_with_indexer4")
+{
+    CheckResult result = check(R"(
+        local function foo(a: {[string]: number, a: string}, i: string)
+            return a[i]
+        end
+        local hi: number = foo({ a = "hi" }, "a") -- shouldn't typecheck since at runtime hi is "hi"
+    )");
+
+    // This typechecks but shouldn't
+    LUAU_REQUIRE_NO_ERRORS(result);
 }
 
 TEST_CASE_FIXTURE(Fixture, "table_subtyping_with_missing_props_dont_report_multiple_errors")
@@ -1446,22 +1536,32 @@ TEST_CASE_FIXTURE(Fixture, "table_subtyping_with_missing_props_dont_report_multi
 TEST_CASE_FIXTURE(Fixture, "table_subtyping_with_extra_props_dont_report_multiple_errors")
 {
     CheckResult result = check(R"(
-        local vec3 = {x = 1, y = 2, z = 3}
-        local vec1 = {x = 1}
+        local vec3 = {{x = 1, y = 2, z = 3}}
+        local vec1 = {{x = 1}}
 
         vec1 = vec3
     )");
 
     LUAU_REQUIRE_ERROR_COUNT(1, result);
 
-    MissingProperties* mp = get<MissingProperties>(result.errors[0]);
-    REQUIRE(mp);
-    CHECK_EQ(mp->context, MissingProperties::Extra);
-    REQUIRE_EQ(2, mp->properties.size());
-    CHECK_EQ(mp->properties[0], "y");
-    CHECK_EQ(mp->properties[1], "z");
-    CHECK_EQ("vec1", toString(mp->superType));
-    CHECK_EQ("vec3", toString(mp->subType));
+    TypeMismatch* tm = get<TypeMismatch>(result.errors[0]);
+    REQUIRE(tm);
+    CHECK_EQ("vec1", toString(tm->wantedType));
+    CHECK_EQ("vec3", toString(tm->givenType));
+}
+
+TEST_CASE_FIXTURE(Fixture, "table_subtyping_with_extra_props_is_ok")
+{
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
+    CheckResult result = check(R"(
+        local vec3 = {x = 1, y = 2, z = 3}
+        local vec1 = {x = 1}
+
+        vec1 = vec3
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
 }
 
 TEST_CASE_FIXTURE(Fixture, "type_mismatch_on_massive_table_is_cut_short")
@@ -1824,4 +1924,199 @@ TEST_CASE_FIXTURE(Fixture, "invariant_table_properties_means_instantiating_table
     LUAU_REQUIRE_ERRORS(result);
 }
 
+TEST_CASE_FIXTURE(Fixture, "table_insert_should_cope_with_optional_properties_in_nonstrict")
+{
+    CheckResult result = check(R"(
+        --!nonstrict
+        local buttons = {}
+        table.insert(buttons, { a = 1 })
+        table.insert(buttons, { a = 2, b = true })
+        table.insert(buttons, { a = 3 })
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "table_insert_should_cope_with_optional_properties_in_strict")
+{
+    ScopedFastFlag sff{"LuauTableSubtypingVariance2", true};
+
+    CheckResult result = check(R"(
+        --!strict
+        local buttons = {}
+        table.insert(buttons, { a = 1 })
+        table.insert(buttons, { a = 2, b = true })
+        table.insert(buttons, { a = 3 })
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "error_detailed_prop")
+{
+    ScopedFastFlag LuauTableSubtypingVariance2{"LuauTableSubtypingVariance2", true}; // Only for new path
+    ScopedFastFlag luauExtendedTypeMismatchError{"LuauExtendedTypeMismatchError", true};
+
+    CheckResult result = check(R"(
+type A = { x: number, y: number }
+type B = { x: number, y: string }
+
+local a: A
+local b: B = a
+    )");
+
+    LUAU_REQUIRE_ERRORS(result);
+    CHECK_EQ(toString(result.errors[0]), R"(Type 'A' could not be converted into 'B'
+caused by:
+  Property 'y' is not compatible. Type 'number' could not be converted into 'string')");
+}
+
+TEST_CASE_FIXTURE(Fixture, "error_detailed_prop_nested")
+{
+    ScopedFastFlag LuauTableSubtypingVariance2{"LuauTableSubtypingVariance2", true}; // Only for new path
+    ScopedFastFlag luauExtendedTypeMismatchError{"LuauExtendedTypeMismatchError", true};
+
+    CheckResult result = check(R"(
+type AS = { x: number, y: number }
+type BS = { x: number, y: string }
+
+type A = { a: boolean, b: AS }
+type B = { a: boolean, b: BS }
+
+local a: A
+local b: B = a
+    )");
+
+    LUAU_REQUIRE_ERRORS(result);
+    CHECK_EQ(toString(result.errors[0]), R"(Type 'A' could not be converted into 'B'
+caused by:
+  Property 'b' is not compatible. Type 'AS' could not be converted into 'BS'
+caused by:
+  Property 'y' is not compatible. Type 'number' could not be converted into 'string')");
+}
+
+TEST_CASE_FIXTURE(Fixture, "error_detailed_metatable_prop")
+{
+    ScopedFastFlag LuauTableSubtypingVariance2{"LuauTableSubtypingVariance2", true}; // Only for new path
+    ScopedFastFlag luauExtendedTypeMismatchError{"LuauExtendedTypeMismatchError", true};
+
+    CheckResult result = check(R"(
+local a1 = setmetatable({ x = 2, y = 3 }, { __call = function(s) end });
+local b1 = setmetatable({ x = 2, y = "hello" }, { __call = function(s) end });
+local c1: typeof(a1) = b1
+
+local a2 = setmetatable({ x = 2, y = 3 }, { __call = function(s) end });
+local b2 = setmetatable({ x = 2, y = 4 }, { __call = function(s, t) end });
+local c2: typeof(a2) = b2
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(2, result);
+    CHECK_EQ(toString(result.errors[0]), R"(Type 'b1' could not be converted into 'a1'
+caused by:
+  Type '{| x: number, y: string |}' could not be converted into '{| x: number, y: number |}'
+caused by:
+  Property 'y' is not compatible. Type 'string' could not be converted into 'number')");
+
+    if (FFlag::LuauExtendedFunctionMismatchError)
+    {
+        CHECK_EQ(toString(result.errors[1]), R"(Type 'b2' could not be converted into 'a2'
+caused by:
+  Type '{| __call: (a, b) -> () |}' could not be converted into '{| __call: <a>(a) -> () |}'
+caused by:
+  Property '__call' is not compatible. Type '(a, b) -> ()' could not be converted into '<a>(a) -> ()'; different number of generic type parameters)");
+    }
+    else
+    {
+        CHECK_EQ(toString(result.errors[1]), R"(Type 'b2' could not be converted into 'a2'
+caused by:
+  Type '{| __call: (a, b) -> () |}' could not be converted into '{| __call: <a>(a) -> () |}'
+caused by:
+  Property '__call' is not compatible. Type '(a, b) -> ()' could not be converted into '<a>(a) -> ()')");
+    }
+}
+
+TEST_CASE_FIXTURE(Fixture, "explicitly_typed_table")
+{
+    ScopedFastFlag sffs[] {
+        {"LuauPropertiesGetExpectedType", true},
+        {"LuauExpectedTypesOfProperties", true},
+        {"LuauTableSubtypingVariance2", true},
+    };
+
+    CheckResult result = check(R"(
+--!strict
+type Super = { x : number }
+type Sub = { x : number, y: number }
+type HasSuper = { p : Super }
+type HasSub = { p : Sub }
+local a: HasSuper = { p = { x = 5, y = 7 }}
+a.p = { x = 9 }
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "explicitly_typed_table_error")
+{
+    ScopedFastFlag sffs[] {
+        {"LuauPropertiesGetExpectedType", true},
+        {"LuauExpectedTypesOfProperties", true},
+        {"LuauTableSubtypingVariance2", true},
+        {"LuauExtendedTypeMismatchError", true},
+    };
+
+    CheckResult result = check(R"(
+--!strict
+type Super = { x : number }
+type Sub = { x : number, y: number }
+type HasSuper = { p : Super }
+type HasSub = { p : Sub }
+local tmp = { p = { x = 5, y = 7 }}
+local a: HasSuper = tmp
+a.p = { x = 9 }
+-- needs to be an error because
+local y: number = tmp.p.y
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ(toString(result.errors[0]), R"(Type 'tmp' could not be converted into 'HasSuper'
+caused by:
+  Property 'p' is not compatible. Table type '{| x: number, y: number |}' not compatible with type 'Super' because the former has extra field 'y')");
+}
+
+TEST_CASE_FIXTURE(Fixture, "explicitly_typed_table_with_indexer")
+{
+    ScopedFastFlag sffs[] {
+        {"LuauPropertiesGetExpectedType", true},
+        {"LuauExpectedTypesOfProperties", true},
+        {"LuauTableSubtypingVariance2", true},
+    };
+
+    CheckResult result = check(R"(
+--!strict
+type Super = { x : number }
+type Sub = { x : number, y: number }
+type HasSuper = { [string] : Super }
+type HasSub = { [string] : Sub }
+local a: HasSuper = { p = { x = 5, y = 7 }}
+a.p = { x = 9 }
+    )");
+
+    LUAU_REQUIRE_NO_ERRORS(result);
+}
+
+TEST_CASE_FIXTURE(Fixture, "recursive_metatable_type_call")
+{
+    ScopedFastFlag luauFixRecursiveMetatableCall{"LuauFixRecursiveMetatableCall", true};
+
+    CheckResult result = check(R"(
+local b
+b = setmetatable({}, {__call = b})
+b()
+    )");
+
+    LUAU_REQUIRE_ERROR_COUNT(1, result);
+    CHECK_EQ(toString(result.errors[0]), R"(Cannot call non-function t1 where t1 = { @metatable {| __call: t1 |}, {  } })");
+}
+
 TEST_SUITE_END();
